# Copyright 2019 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
r"""
This module defines and implements the Python-embedded quantum programming language
for continuous-variable (CV) quantum systems.
The syntax is modeled after ProjectQ :cite:`projectq2016`.
"""
from collections.abc import Sequence
import copy
import warnings

import numpy as np

from scipy.linalg import block_diag
import scipy.special as ssp

import strawberryfields as sf
import strawberryfields.program_utils as pu
import strawberryfields.decompositions as dec
from .backends.states import BaseFockState, BaseGaussianState
from .backends.shared_ops import changebasis
from .program_utils import Command, RegRef, MergeFailure
from .parameters import par_regref_deps, par_str, par_evaluate, par_is_symbolic, par_funcs as pf

# pylint: disable=abstract-method
# pylint: disable=protected-access
# pylint: disable=arguments-differ  # Measurement._apply introduces the "shots" argument

# numerical tolerances
_decomposition_merge_tol = 1e-13
_decomposition_tol = (
    1e-13  # TODO this tolerance is used for various purposes and is not well-defined
)


def warning_on_one_line(message, category, filename, lineno, file=None, line=None):
    """User warning formatter"""
    # pylint: disable=unused-argument
    return "{}:{}: {}: {}\n".format(filename, lineno, category.__name__, message)


warnings.formatwarning = warning_on_one_line


def _seq_to_list(s):
    "Converts a Sequence or a single object into a list."
    if not isinstance(s, Sequence):
        s = [s]
    return list(s)


class Operation:
    """Abstract base class for quantum operations acting on one or more subsystems.

    :attr:`Operation.measurement_deps` is a set containing the :class:`.RegRef`
    the :class:`Operation` depends on through its parameters.
    In the quantum circuit diagram notation it corresponds to the vertical double lines of classical
    information entering the :class:`Operation` that originate in the measurement of a subsystem.

    This abstract base class may be initialised with parameters; see the
    :class:`~strawberryfields.parameters.Parameter` class for more details.

    Args:
        par (Sequence[Any]): Operation parameters. An empty sequence if no parameters
            are required.
    """

    # default: one-subsystem operation
    #: int: number of subsystems the operation acts on, or None if any number > 0 is ok
    ns = 1

    def __init__(self, par):
        #: set[RegRef]: extra dependencies due to deferred measurements, used during optimization
        self._measurement_deps = set()
        #: list[Parameter]: operation parameters
        self.p = []

        # convert each parameter into a Parameter instance, keep track of dependenciens
        for q in par:
            if isinstance(q, RegRef):
                raise TypeError("Use RegRef.par for measured parameters.")
            self.p.append(q)
            self._measurement_deps |= par_regref_deps(q)

    def __str__(self):
        """String representation for the Operation using Blackbird syntax.

        Returns:
            str: string representation
        """
        # defaults to the class name
        if not self.p:
            return self.__class__.__name__

        # class name and parameter values
        temp = [par_str(i) for i in self.p]
        return self.__class__.__name__ + "(" + ", ".join(temp) + ")"

    @property
    def measurement_deps(self):
        """Extra dependencies due to parameters that depend on measurements.

        Returns:
            set[RegRef]: dependencies
        """
        return self._measurement_deps

    def __or__(self, reg):
        """Apply the operation to a part of a quantum register.

        Appends the Operation to a :class:`.Program` instance.

        Args:
            reg (RegRef, Sequence[RegRef]): subsystem(s) the operation is acting on

        Returns:
            list[RegRef]: subsystem list as RegRefs
        """
        # into a list of subsystems
        reg = _seq_to_list(reg)
        if (not reg) or (self.ns is not None and self.ns != len(reg)):
            raise ValueError("Wrong number of subsystems.")
        # append it to the Program
        reg = pu.Program_current_context.append(self, reg)
        return reg

    def merge(self, other):
        """Merge the operation with another (acting on the exact same set of subsystems).

        .. note:: For subclass overrides: merge may return a newly created object,
           or self, or other, but it must never modify self or other
           because the same Operation objects may be also used elsewhere.

        Args:
            other (Operation): operation to merge this one with

        Returns:
            Operation, None: other * self. The return value None represents
            the identity gate (doing nothing).

        Raises:
            .MergeFailure: if the two operations cannot be merged
        """
        # todo: Using the return value None to denote the identity is a
        # bit dangerous, since a function with no explicit return statement
        # also returns None, which can lead to puzzling bugs. Maybe return
        # a special singleton Identity object instead?
        raise NotImplementedError

    def decompose(self, reg, **kwargs):
        """Decompose the operation into elementary operations supported by the backend API.

        See :mod:`strawberryfields.backends.base`.

        Args:
            reg (Sequence[RegRef]): subsystems the operation is acting on

        Returns:
            list[Command]: decomposition as a list of operations acting on specific subsystems
        """
        return self._decompose(reg, **kwargs)

    def _decompose(self, reg, **kwargs):
        """Internal decomposition method defined by subclasses.

        NOTE: Does not evaluate Operation parameters, symbolic parameters remain symbolic.

        Args:
            reg (Sequence[RegRef]): subsystems the operation is acting on

        Returns:
            list[Command]: decomposition as a list of operations acting on specific subsystems
        """
        raise NotImplementedError("No decomposition available: {}".format(self))

    def _apply(self, reg, backend, **kwargs):
        """Internal apply method. Uses numeric subsystem referencing.

        Args:
            reg (Sequence[int]): subsystem indices the operation is
                acting on (this is how the backend API wants them)
            backend (BaseBackend): backend to execute the operation

        Returns:
            array[Number] or None: Measurement results, if any; shape == (len(reg), shots).
        """
        raise NotImplementedError("Missing direct implementation: {}".format(self))

    def apply(self, reg, backend, **kwargs):
        """Ask a local backend to execute the operation on the current register state right away.

        Takes care of parameter evaluations and any pending formal
        transformations (like dagger) and then calls :meth:`Operation._apply`.

        Args:
            reg (Sequence[RegRef]): subsystem(s) the operation is acting on
            backend (BaseBackend): backend to execute the operation

        Returns:
            Any: the result of self._apply
        """
        # NOTE: We cannot just replace all parameters with their evaluated
        # numerical values here. If we re-initialize a measured mode and
        # re-measure it, the corresponding MeasuredParameter value should change accordingly
        # when it is used again after the new measurement.

        # convert RegRefs back to indices for the backend API
        temp = [rr.ind for rr in reg]
        # call the child class specialized _apply method
        return self._apply(temp, backend, **kwargs)


# ====================================================================
# Derived operation classes
# ====================================================================


class Preparation(Operation):
    """Abstract base class for operations that demolish
    the previous state of the subsystem entirely.
    """

    def merge(self, other):
        # sequential preparation, only the last one matters
        if isinstance(other, Preparation):
            # give a warning, since this is pointless and probably a user error
            warnings.warn("Two subsequent state preparations, first one has no effect.")
            return other

        raise MergeFailure("For now, Preparations cannot be merged with anything else.")


class Measurement(Operation):
    """Abstract base class for subsystem measurements.

    The measurement is deferred: its result is available only
    after the backend has executed it. The value of the measurement can
    be accessed in the program through the symbolic subsystem reference
    to the measured subsystem.

    When the measurement happens, the state of the system is updated
    to the conditional state corresponding to the measurement result.
    Measurements also support postselection, see below.

    Args:
        select (None, Sequence[Number]): Desired values of the measurement
            results, one for each subsystem the measurement acts on.
            Allows the post-selection of specific measurement results
            instead of randomly sampling. None means no postselection.
    """

    # todo: self.select could support :class:`~strawberryfields.parameters.Parameter` instances.
    ns = None

    def __init__(self, par, select=None):
        super().__init__(par)
        #: None, Sequence[Number]: postselection values, one for each measured subsystem
        self.select = select

    def __str__(self):
        # class name, parameter values, and possibly the select parameter
        temp = super().__str__()
        if self.select is not None:
            temp = temp[:-1] + ", select={})".format(self.select)
        return temp

    def merge(self, other):
        raise MergeFailure("For now, measurements cannot be merged with anything else.")

    def apply(self, reg, backend, **kwargs):
        """Ask a backend to execute the operation on the current register state right away.

        Like :func:`Operation.apply`, but also stores the measurement result in the RegRefs.

        Keyword Args:
            shots (int): Number of independent evaluations to perform.
                Only applies to Measurements.
        """
        values = super().apply(reg, backend, **kwargs)
        # convert the returned values into an iterable with the measured modes indexed along
        # the first axis and shots along second axis (if larger than 1), so that we can assign
        # register values
        shots = kwargs.get("shots", 1)
        if self.ns == 1:
            values = [values]  # values is either a scalar, or has shape (shots,)
        else:
            if shots > 1:
                values = values.T  # shape of values would be (shots, num_meas,)

        # store the results in the register reference objects
        for v, r in zip(values, reg):
            r.val = v


class Decomposition(Operation):
    """Abstract base class for multimode matrix transformations.

    This class provides the base behaviour for decomposing various multimode operations
    into a sequence of gates and state preparations.

    .. note:: The first parameter ``p[0]`` of a Decomposition is always a square matrix, and it cannot be symbolic.
    """

    ns = None  # overridden by child classes in __init__

    @staticmethod
    def _check_p0(p0):
        """Checks that p0 is not symbolic."""
        if par_is_symbolic(p0):
            raise TypeError(
                "The first parameter of a Decomposition is a square matrix, and cannot be symbolic."
            )

    def __init__(self, par, decomp=True):
        self._check_p0(par[0])
        super().__init__(par)
        self.decomp = decomp
        """bool: If False, try to apply the Decomposition as a single primitive operation
        instead of decomposing it."""

    def merge(self, other):
        # can be merged if they are the same class
        if isinstance(other, self.__class__):
            # at the moment, we will assume all state decompositions only
            # take one argument. The only exception currently are state
            # decompositions, which cannot be merged.
            U1 = self.p[0]
            U2 = other.p[0]
            U = U2 @ U1
            # Note: above we strip the Parameter wrapper to make the following check
            # easier to perform. The constructor restores it.
            # Another option would be to add the required methods to Parameter class.
            # check if the matrices cancel
            if np.allclose(U, np.identity(len(U)), atol=_decomposition_merge_tol, rtol=0):
                return None

            return self.__class__(U)

        raise MergeFailure("Not the same decomposition type.")


class Transformation(Operation):
    """Abstract base class for transformations.

    This class provides the base behaviour for operations which
    act on existing states.
    """

    # NOTE: At the moment this is an empty class, and only
    # exists for a nicer inheritence diagram. One option is
    # to remove, and make Channel and Gate top-level derived classes.
    #
    # Are there any useful operations/properties shared by Gate/Channel?


# ====================================================================
# Derived transformation classes
# ====================================================================


class Channel(Transformation):
    """Abstract base class for quantum channels.

    This class provides the base behaviour for non-unitary
    maps and transformations.
    """

    # TODO decide how all Channels should treat the first parameter p[0]
    # (see e.g. https://en.wikipedia.org/wiki/C0-semigroup), cf. p[0] in ops.Gate

    def merge(self, other):
        if not self.__class__ == other.__class__:
            raise MergeFailure("Not the same channel family.")

        # channels can be merged if they are the same class and share all the other parameters
        if self.p[1:] == other.p[1:]:
            # determine the combined first parameter
            T = self.p[0] * other.p[0]
            # if one, replace with the identity
            if T == 1:
                return None

            # return a copy
            # NOTE deepcopy would make copies of the parameters which would mess things up
            temp = copy.copy(self)
            temp.p = [T] + self.p[1:]  # change the parameter list
            return temp

        raise MergeFailure("Don't know how to merge these operations.")


class Gate(Transformation):
    """Abstract base class for unitary quantum gates.

    The first parameter p[0] of the Gate class is special:

    * The value p[0] = 0 corresponds to the identity gate.
    * The inverse gate is obtained by negating p[0].
    * Two gates of this class can be merged by adding the
      first parameters together, assuming all the other parameters match.
    """

    def __init__(self, par):
        super().__init__(par)
        # default: non-dagger form
        self.dagger = False  #: bool: formal inversion of the gate

    def __str__(self):
        """String representation for the gate."""
        # add a dagger symbol to the class name if needed
        temp = super().__str__()
        if self.dagger:
            temp += ".H"
        return temp

    @property
    def H(self):
        """Returns a copy of the gate with the self.dagger flag flipped.

        H stands for hermitian conjugate.

        Returns:
            Gate: formal inverse of this gate
        """
        # HACK Semantically a bad use of @property since this method is not a getter.
        # NOTE deepcopy would make copies of the parameters which would mess things up
        s = copy.copy(self)
        s.dagger = not s.dagger
        return s

    def decompose(self, reg, **kwargs):
        """Decompose the operation into elementary operations supported by the backend API.

        Like :func:`Operation.decompose`, but applies self.dagger.
        """
        seq = self._decompose(reg, **kwargs)
        if self.dagger:
            # apply daggers, reverse the Command sequence
            for cmd in seq:
                cmd.op.dagger = not cmd.op.dagger
            seq = list(reversed(seq))
        return seq

    def apply(self, reg, backend, **kwargs):
        """Ask a backend to execute the operation on the current register state right away.

        Like :func:`Operation.apply`, but takes into account the special nature of
        p[0] and applies self.dagger.

        Returns:
            None: Gates do not return anything, return value is None
        """
        z = self.p[0]
        # if z represents a batch of parameters, then all of these
        # must be zero to skip calling backend
        if np.all(z == 0):
            # identity, no need to apply
            return
        if self.dagger:
            z = -z
        original_p0 = self.p[0]  # store the original Parameter
        self.p[0] = z

        # convert RegRefs back to indices for the backend API
        temp = [rr.ind for rr in reg]
        # call the child class specialized _apply method
        self._apply(temp, backend, **kwargs)
        self.p[0] = original_p0  # restore the original Parameter instance

    def merge(self, other):
        if not self.__class__ == other.__class__:
            raise MergeFailure("Not the same gate family.")

        # gates can be merged if they are the same class and share all the other parameters
        if self.p[1:] == other.p[1:]:
            # make sure the gates have the same dagger flag, if not, invert the second p[0]
            if self.dagger == other.dagger:
                temp = other.p[0]
            else:
                temp = -other.p[0]
            # now we can add up the parameters and keep self.dagger
            p0 = self.p[0] + temp
            if p0 == 0:
                return None  # identity gate

            # return a copy
            # NOTE deepcopy would make copies the parameters which would mess things up
            temp = copy.copy(self)
            temp.p = [p0] + self.p[1:]  # change the parameter list
            return temp

        raise MergeFailure("Don't know how to merge these gates.")


# ====================================================================
# State preparation operations
# ====================================================================


class Vacuum(Preparation):
    """Prepare a mode in the :ref:`vacuum state <vacuum_state>`.

    Can be accessed via the shortcut variable ``Vac``.
    """

    def __init__(self):
        super().__init__([])

    def _apply(self, reg, backend, **kwargs):
        backend.prepare_vacuum_state(*reg)

    def __str__(self):
        # return the shorthand object when the
        # command is printed by the user
        return "Vac"


class Coherent(Preparation):
    r"""Prepare a mode in a :ref:`coherent state <coherent_state>`.

    The gate is parameterized so that a user can specify a single complex number :math:`a=\alpha`
    or use the polar form :math:`a = r, p=\phi` and still get the same result.

    Args:
        a (complex): displacement parameter :math:`\alpha`
        p (float): phase angle :math:`\phi`
    """

    def __init__(self, a=0.0, p=0.0):
        super().__init__([a, p])

    def _apply(self, reg, backend, **kwargs):
        p = self.p[0] * pf.exp(1j * self.p[1])
        z = par_evaluate(p)
        backend.prepare_coherent_state(z, *reg)


class Squeezed(Preparation):
    r"""Prepare a mode in a :ref:`squeezed vacuum state <squeezed_state>`.

    Args:
        r (float): squeezing magnitude
        p (float): squeezing angle :math:`\phi`
    """

    def __init__(self, r=0.0, p=0.0):
        super().__init__([r, p])

    def _apply(self, reg, backend, **kwargs):
        p = par_evaluate(self.p)
        backend.prepare_squeezed_state(p[0], p[1], *reg)


class DisplacedSqueezed(Preparation):
    r"""Prepare a mode in a :ref:`displaced squeezed state <displaced_squeezed_state>`.

    A displaced squeezed state is prepared by squeezing a vacuum state, and
    then applying a displacement operator.

    .. math::
       \ket{\alpha,z} = D(\alpha)\ket{0,z} = D(\alpha)S(z)\ket{0},

    where the squeezing parameter :math:`z=re^{i\phi}`.

    Args:
        alpha (complex): displacement parameter
        r (float): squeezing magnitude
        p (float): squeezing angle :math:`\phi`
    """

    def __init__(self, alpha=0.0, r=0.0, p=0.0):
        super().__init__([alpha, r, p])

    def _apply(self, reg, backend, **kwargs):
        p = par_evaluate(self.p)
        # prepare the displaced squeezed state directly
        backend.prepare_displaced_squeezed_state(p[0], p[1], p[2], *reg)

    def _decompose(self, reg, **kwargs):
        # squeezed state preparation followed by a displacement gate
        return [Command(Squeezed(self.p[1], self.p[2]), reg), Command(Dgate(self.p[0]), reg)]


class Fock(Preparation):
    r"""Prepare a mode in a :ref:`fock_basis` state.

    The prepared mode is traced out and replaced with the Fock state :math:`\ket{n}`.
    As a result the state of the other subsystems may have to be described using a density matrix.

    Args:
        n (int): Fock state to prepare
    """

    def __init__(self, n=0):
        super().__init__([n])

    def _apply(self, reg, backend, **kwargs):
        p = par_evaluate(self.p)
        backend.prepare_fock_state(p[0], *reg)


class Catstate(Preparation):
    r"""Prepare a mode in a :ref:`cat state <cat_state>`.

    A cat state is the coherent superposition of two coherent states,

    .. math::
       \ket{\text{cat}(\alpha)} = \frac{1}{N} (\ket{\alpha} +e^{i\phi} \ket{-\alpha}),

    where :math:`N = \sqrt{2 (1+\cos(\phi)e^{-2|\alpha|^2})}` is the normalization factor.

    Args:
        alpha (complex): displacement parameter
        p (float): parity, where :math:`\phi=p\pi`. ``p=0`` corresponds to an even
            cat state, and ``p=1`` an odd cat state.
    """

    def __init__(self, alpha=0, p=0):
        super().__init__([alpha, p])

    def _apply(self, reg, backend, **kwargs):
        alpha = self.p[0]
        phi = np.pi * self.p[1]
        D = backend.get_cutoff_dim()
        l = np.arange(D)[:, np.newaxis]

        # normalization constant
        temp = pf.exp(-0.5 * pf.Abs(alpha) ** 2)
        N = temp / pf.sqrt(2 * (1 + pf.cos(phi) * temp ** 4))

        # coherent states
        c1 = (alpha ** l) / np.sqrt(ssp.factorial(l))
        c2 = ((-alpha) ** l) / np.sqrt(ssp.factorial(l))
        # add them up with a relative phase
        ket = (c1 + pf.exp(1j * phi) * c2) * N

        # in order to support broadcasting, the batch axis has been located at last axis, but backend expects it up as first axis
        ket = np.transpose(ket)

        # drop dummy batch axis if it is not necessary
        ket = np.squeeze(ket)

        # evaluate the array (elementwise)
        ket = par_evaluate(ket)
        backend.prepare_ket_state(ket, *reg)


class Ket(Preparation):
    r"""Prepare mode(s) using the given ket vector(s) in the Fock basis.

    The prepared modes are traced out and replaced with the given ket state
    (in the Fock basis). As a result the state of the other subsystems may have
    to be described using a density matrix.

    The provided kets must be each be of length ``cutoff_dim``, matching
    the cutoff dimension used in calls to :meth:`eng.run <~.Engine.run>`.

    Args:
        state (array or BaseFockState): state vector in the Fock basis.
            This can be provided as either:

            * a single ket vector, for single mode state preparation
            * a multimode ket, with one array dimension per mode
            * a :class:`BaseFockState` state object.
    """
    ns = None

    def __init__(self, state):
        if isinstance(state, BaseFockState):
            if not state.is_pure:
                raise ValueError("Provided Fock state is not pure.")
            super().__init__([state.ket()])
        elif isinstance(state, BaseGaussianState):
            raise ValueError("Gaussian states are not supported for the Ket operation.")
        else:
            super().__init__([state])

    def _apply(self, reg, backend, **kwargs):
        p = par_evaluate(self.p)
        backend.prepare_ket_state(p[0], reg)


class DensityMatrix(Preparation):
    r"""Prepare mode(s) using the given density matrix in the Fock basis.

    The prepared modes are traced out and replaced with the given state
    (in the Fock basis). As a result, the overall state of system
    will also have to be described using a density matrix.

    The provided density matrices must be of size ``[cutoff_dim, cutoff_dim]``,
    matching the cutoff dimension used in calls to :meth:`eng.run <~.Engine.run>`.

    Args:
        state (array or BaseFockState): density matrix in the Fock basis.
            This can be provided as either:

            * a single mode two-dimensional matrix :math:`\rho_{ij}`,
            * a multimode tensor :math:`\rho_{ij,kl,\dots,mn}`, with two indices per mode,
            * a :class:`BaseFockState` state object.
    """
    ns = None

    def __init__(self, state):
        if isinstance(state, BaseFockState):
            super().__init__([state.dm()])
        elif isinstance(state, BaseGaussianState):
            raise ValueError("Gaussian states are not supported for the Ket operation.")
        else:
            super().__init__([state])

    def _apply(self, reg, backend, **kwargs):
        p = par_evaluate(self.p)
        backend.prepare_dm_state(p[0], reg)


class Thermal(Preparation):
    r"""Prepare a mode in a :ref:`thermal state <thermal_state>`.

    The requested mode is traced out and replaced with the thermal state :math:`\rho(\bar{n})`.
    As a result the state will be described using a density matrix.

    Args:
        n (float): mean thermal population of the mode
    """

    def __init__(self, n=0):
        super().__init__([n])

    def _apply(self, reg, backend, **kwargs):
        p = par_evaluate(self.p)
        backend.prepare_thermal_state(p[0], *reg)


# ====================================================================
# Measurements
# ====================================================================


class MeasureFock(Measurement):
    """:ref:`photon_counting`: measures a set of modes in the Fock basis.

    Also accessible via the shortcut variable ``Measure``.

    After measurement, the modes are reset to the vacuum state.
    """

    ns = None

    def __init__(self, select=None):
        if select is not None and not isinstance(select, Sequence):
            select = [select]
        super().__init__([], select)

    def _apply(self, reg, backend, shots=1, **kwargs):
        return backend.measure_fock(reg, shots=shots, select=self.select, **kwargs)


class MeasureThreshold(Measurement):
    """Measures a set of modes with thresholded Fock-state measurements, i.e.,
    measuring whether a mode contain zero or nonzero photons.

    After measurement, the modes are reset to the vacuum state.
    """

    ns = None

    def __init__(self, select=None):
        if select is not None and not isinstance(select, Sequence):
            select = [select]
        super().__init__([], select)

    def _apply(self, reg, backend, shots=1, **kwargs):
        return backend.measure_threshold(reg, shots=shots, select=self.select, **kwargs)


class MeasureHomodyne(Measurement):
    r"""Performs a :ref:`homodyne measurement <homodyne>`, measures one quadrature of a mode.

    * Position basis measurement: :math:`\phi = 0`
      (also accessible via the shortcut variable ``MeasureX``).

    * Momentum basis measurement: :math:`\phi = \pi/2`.
      (also accessible via the shortcut variable ``MeasureP``)

    The measured mode is reset to the vacuum state.

    Args:
        phi (float): measurement angle :math:`\phi`
        select (None, float): (Optional) desired values of measurement result.
            Allows the post-selection of specific measurement results instead of randomly sampling.
    """
    ns = 1

    def __init__(self, phi, select=None):
        super().__init__([phi], select)

    def _apply(self, reg, backend, shots=1, **kwargs):
        p = par_evaluate(self.p)
        s = np.sqrt(sf.hbar / 2)  # scaling factor, since the backend API call is hbar-independent
        select = self.select
        if select is not None:
            select = select / s

        return s * backend.measure_homodyne(p[0], *reg, shots=shots, select=select, **kwargs)

    def __str__(self):
        if self.select is None:
            if self.p[0] == 0:
                return "MeasureX"
            if self.p[0] == np.pi / 2:
                return "MeasureP"
        return super().__str__()


class MeasureHeterodyne(Measurement):
    r"""Performs a :ref:`heterodyne measurement <heterodyne>` on a mode.

    Also accessible via the shortcut variable ``MeasureHD``.

    Samples the joint Husimi distribution :math:`Q(\vec{\alpha}) = \frac{1}{\pi}\bra{\vec{\alpha}}\rho\ket{\vec{\alpha}}`.
    The measured mode is reset to the vacuum state.

    Args:
        select (None, complex): (Optional) desired values of measurement result.
            Allows the post-selection of specific measurement results instead of randomly sampling.
    """
    ns = 1

    def __init__(self, select=None):
        super().__init__([], select)

    def _apply(self, reg, backend, shots=1, **kwargs):
        return backend.measure_heterodyne(*reg, shots=shots, select=self.select, **kwargs)

    def __str__(self):
        if self.select is None:
            return "MeasureHD"
        return "MeasureHeterodyne(select={})".format(self.select)


# ====================================================================
# Channels
# ====================================================================


class LossChannel(Channel):
    r"""Perform a :ref:`loss channel <loss>` operation on the specified mode.

    This channel couples mode :math:`\a` to another bosonic mode :math:`\hat{b}`
    prepared in the vacuum state using the following transformation:

    .. math::
       \a \mapsto \sqrt{T} a+\sqrt{1-T} \hat{b}

    Args:
        T (float): the loss parameter :math:`0\leq T\leq 1`.
    """

    def __init__(self, T):
        super().__init__([T])

    def _apply(self, reg, backend, **kwargs):
        p = par_evaluate(self.p)
        backend.loss(p[0], *reg)


class ThermalLossChannel(Channel):
    r"""Perform a :ref:`thermal loss channel <thermal_loss>` operation on the specified mode.

    This channel couples mode :math:`\a` to another bosonic mode :math:`\hat{b}`
    prepared in a thermal state with mean photon number :math:`\bar{n}`,
    using the following transformation:

    .. math::
       \a \mapsto \sqrt{T} a+\sqrt{1-T} \hat{b}

    Args:
        T (float): the loss parameter :math:`0\leq T\leq 1`.
        nbar (float): mean photon number of the environment thermal state
    """

    def __init__(self, T, nbar):
        super().__init__([T, nbar])

    def _apply(self, reg, backend, **kwargs):
        p = par_evaluate(self.p)
        backend.thermal_loss(p[0], p[1], *reg)


# ====================================================================
# Unitary gates
# ====================================================================


class Dgate(Gate):
    r"""Phase space :ref:`displacement <displacement>` gate.

    .. math::
       D(\alpha) = \exp(\alpha a^\dagger -\alpha^* a) = \exp\left(-i\sqrt{2}(\re(\alpha) \hat{p} -\im(\alpha) \hat{x})/\sqrt{\hbar}\right)

    where :math:`\alpha = r e^{i\phi}` has magnitude :math:`r\geq 0` and phase :math:`\phi`.

    The gate is parameterized so that a user can specify a single complex number :math:`a=\alpha`
    or use the polar form :math:`a = r, \phi` and still get the same result.

    Args:
        a (complex): displacement parameter :math:`\alpha`
        phi (float): extra (optional) phase angle :math:`\phi`
    """

    def __init__(self, a, phi=0.0):
        super().__init__([a, phi])

    def _apply(self, reg, backend, **kwargs):
        p = self.p[0] * pf.exp(1j * self.p[1])
        z = par_evaluate(p)
        backend.displacement(z, *reg)


class Xgate(Gate):
    r"""Position :ref:`displacement <displacement>` gate.

    .. math::
       X(x) = e^{-i x \hat{p}/\hbar}

    Args:
        x (float): position displacement
    """

    def __init__(self, x):
        super().__init__([x])

    def _decompose(self, reg, **kwargs):
        # into a displacement
        z = self.p[0] / np.sqrt(2 * sf.hbar)
        return [Command(Dgate(z, 0), reg)]


class Zgate(Gate):
    r"""Momentum :ref:`displacement <displacement>` gate.

    .. math::
       Z(p) = e^{i p \hat{x}/\hbar}

    Args:
        p (float): momentum displacement
    """

    def __init__(self, p):
        super().__init__([p])

    def _decompose(self, reg, **kwargs):
        # into a displacement
        z = self.p[0] * 1j / np.sqrt(2 * sf.hbar)
        return [Command(Dgate(z, 0), reg)]


class Sgate(Gate):
    r"""Phase space :ref:`squeezing <squeezing>` gate.

    .. math::
       S(z) = \exp\left(\frac{1}{2}(z^* a^2 -z {a^\dagger}^2)\right)

    where :math:`z = r e^{i\phi}`.

    Args:
        r (float): squeezing amount
        phi (float): squeezing phase angle :math:`\phi`
    """

    def __init__(self, r, phi=0.0):
        super().__init__([r, phi])

    def _apply(self, reg, backend, **kwargs):
        p = self.p[0] * pf.exp(1j * self.p[1])
        z = par_evaluate(p)
        backend.squeeze(z, *reg)


class Pgate(Gate):
    r""":ref:`Quadratic phase <quadratic>` gate.

    .. math::
       P(s) = e^{i \frac{s}{2} \hat{x}^2/\hbar}

    Args:
        s (float): parameter
    """

    def __init__(self, s):
        super().__init__([s])

    def _decompose(self, reg, **kwargs):
        # into a squeeze and a rotation
        temp = self.p[0] / 2
        r = pf.acosh(pf.sqrt(1 + temp ** 2))
        theta = pf.atan(temp)
        phi = -np.pi / 2 * pf.sign(temp) - theta
        return [Command(Sgate(r, phi), reg), Command(Rgate(theta), reg)]


class Vgate(Gate):
    r""":ref:`Cubic phase <cubic>` gate.

    .. math::
       V(\gamma) = e^{i \frac{\gamma}{3 \hbar} \hat{x}^3}

    .. warning:: The cubic phase gate has lower accuracy than the Kerr gate at the same cutoff dimension.

    Args:
        gamma (float): parameter
    """

    def __init__(self, gamma):
        super().__init__([gamma])

    def _apply(self, reg, backend, **kwargs):
        gamma_prime = self.p[0] * np.sqrt(sf.hbar / 2)
        # the backend API call cubic_phase is hbar-independent
        backend.cubic_phase(par_evaluate(gamma_prime), *reg)


class Kgate(Gate):
    r""":ref:`Kerr <kerr>` gate.

    .. math::
       K(\kappa) = e^{i \kappa \hat{n}^2}

    Args:
        kappa (float): parameter
    """

    def __init__(self, kappa):
        super().__init__([kappa])

    def _apply(self, reg, backend, **kwargs):
        p = par_evaluate(self.p)
        backend.kerr_interaction(p[0], *reg)


class Rgate(Gate):
    r""":ref:`Rotation <rotation>` gate.

    .. math::
       R(\theta) = e^{i \theta a^\dagger a}

    Args:
        theta (float): rotation angle :math:`\theta`.
    """

    def __init__(self, theta):
        super().__init__([theta])

    def _apply(self, reg, backend, **kwargs):
        p = par_evaluate(self.p)
        backend.rotation(p[0], *reg)


class BSgate(Gate):
    r"""BSgate(theta=pi/4, phi=0.)
    :ref:`Beamsplitter <beamsplitter>` gate.

    .. math::
       B(\theta,\phi) = \exp\left(\theta (e^{i \phi} a_1 a_2^\dagger -e^{-i \phi} a_1^\dagger a_2) \right)

    Args:
        theta (float): Transmittivity angle :math:`\theta`. The transmission amplitude of the beamsplitter is :math:`t = \cos(\theta)`.
            The value :math:`\theta=\pi/4` gives the 50-50 beamsplitter (default).
        phi (float): Phase angle :math:`\phi`. The reflection amplitude of the beamsplitter is :math:`r = e^{i\phi}\sin(\theta)`.
            The value :math:`\phi = \pi/2` gives the symmetric beamsplitter.
    """
    ns = 2

    def __init__(self, theta=np.pi / 4, phi=0.0):
        # default: 50% beamsplitter
        super().__init__([theta, phi])

    def _apply(self, reg, backend, **kwargs):
        t = pf.cos(self.p[0])
        r = pf.sin(self.p[0]) * pf.exp(1j * self.p[1])
        p = par_evaluate([t, r])
        backend.beamsplitter(*p, *reg)


class MZgate(Gate):
    r"""Mach-Zehnder interferometer.

    .. math::

        \mathrm{MZ}(\phi_{in}, \phi_{ex}) = BS\left(\frac{\pi}{4}, \frac{\pi}{2}\right)
            (R(\phi_{in})\otimes I) BS\left(\frac{\pi}{4}, \frac{\pi}{2}\right)
            (R(\phi_{ex})\otimes I)

    Args:
        phi_in (float): internal phase
        phi_ex (float): external phase
    """
    ns = 2

    def __init__(self, phi_in, phi_ex):
        super().__init__([phi_in, phi_ex])

    def _decompose(self, reg, **kwargs):
        # into local phase shifts and two 50-50 beamsplitters
        return [
            Command(Rgate(self.p[1]), reg[0]),
            Command(BSgate(np.pi / 4, np.pi / 2), reg),
            Command(Rgate(self.p[0]), reg[0]),
            Command(BSgate(np.pi / 4, np.pi / 2), reg),
        ]


class S2gate(Gate):
    r""":ref:`Two-mode squeezing <two_mode_squeezing>` gate.

    .. math::
       S_2(z) = \exp\left(z^* ab -z a^\dagger b^\dagger \right) = \exp\left(r (e^{-i\phi} ab -e^{i\phi} a^\dagger b^\dagger \right)

    where :math:`z = r e^{i\phi}`.

    Args:
        r (float): squeezing amount
        phi (float): squeezing phase angle :math:`\phi`
    """
    ns = 2

    def __init__(self, r, phi=0.0):
        super().__init__([r, phi])

    def _apply(self, reg, backend, **kwargs):
        p = self.p[0] * pf.exp(1j * self.p[1])
        z = par_evaluate(p)
        backend.two_mode_squeeze(z, *reg)

    def _decompose(self, reg, **kwargs):
        # two opposite squeezers sandwiched between 50% beamsplitters
        S = Sgate(self.p[0], self.p[1])
        BS = BSgate(np.pi / 4, 0)
        return [Command(BS, reg), Command(S, reg[0]), Command(S.H, reg[1]), Command(BS.H, reg)]


class CXgate(Gate):
    r""":ref:`Controlled addition or sum <CX>` gate in the position basis.

    .. math::
       \text{CX}(s) = \int dx \ket{x}\bra{x} \otimes D\left({\frac{1}{\sqrt{2\hbar}}}s x\right) = e^{-i s \: \hat{x} \otimes \hat{p}/\hbar}

    In the position basis it maps
    :math:`\ket{x_1, x_2} \mapsto \ket{x_1, s x_1 +x_2}`.

    Args:
        s (float): addition multiplier
    """
    ns = 2

    def __init__(self, s=1):
        super().__init__([s])

    def _decompose(self, reg, **kwargs):
        s = self.p[0]
        r = pf.asinh(-s / 2)
        theta = 0.5 * pf.atan2(-1.0 / pf.cosh(r), -pf.tanh(r))
        return [
            Command(BSgate(theta, 0), reg),
            Command(Sgate(r, 0), reg[0]),
            Command(Sgate(-r, 0), reg[1]),
            Command(BSgate(theta + np.pi / 2, 0), reg),
        ]


class CZgate(Gate):
    r""":ref:`Controlled phase <CZ>` gate in the position basis.

    .. math::
       \text{CZ}(s) =  \iint dx dy \: e^{i sxy/\hbar} \ket{x,y}\bra{x,y} = e^{i s \: \hat{x} \otimes \hat{x}/\hbar}

    In the position basis it maps
    :math:`\ket{x_1, x_2} \mapsto e^{i s x_1 x_2/\hbar} \ket{x_1, x_2}`.

    Args:
        s (float): phase shift multiplier
    """
    ns = 2

    def __init__(self, s=1):
        super().__init__([s])

    def _decompose(self, reg, **kwargs):
        # phase-rotated CZ
        CX = CXgate(self.p[0])
        return [
            Command(Rgate(-np.pi / 2), reg[1]),
            Command(CX, reg),
            Command(Rgate(np.pi / 2), reg[1]),
        ]


class CKgate(Gate):
    r""":ref:`Cross-Kerr <cross_kerr>` gate.

    .. math::
       CK(\kappa) = e^{i \kappa \hat{n}_1\hat{n}_2}

    Args:
        kappa (float): parameter
    """
    ns = 2

    def __init__(self, kappa):
        super().__init__([kappa])

    def _apply(self, reg, backend, **kwargs):
        p = par_evaluate(self.p)
        backend.cross_kerr_interaction(p[0], *reg)


class Fouriergate(Gate):
    r""":ref:`Fourier <fourier>` gate.

    Also accessible via the shortcut variable ``Fourier``.

    A special case of the :class:`phase space rotation gate <Rgate>`, where :math:`\theta=\pi/2`.

    .. math::
       F = R(\pi/2) = e^{i (\pi/2) a^\dagger a}
    """

    def __init__(self):
        super().__init__([np.pi / 2])

    def _decompose(self, reg, **kwargs):
        # into a rotation
        theta = np.pi / 2
        return [Command(Rgate(theta), reg)]

    def __str__(self):
        """String representation for the gate."""
        temp = "Fourier"
        if self.dagger:
            temp += ".H"
        return temp


# ====================================================================
# Metaoperations
# ====================================================================


# ====================================================================
# Subsystem creation and deletion
# ====================================================================


class MetaOperation(Operation):
    """Abstract base class for metaoperations.

    This includes subsystem creation and deletion.
    """

    def __init__(self):
        super().__init__(par=[])


class _Delete(MetaOperation):
    """Deletes one or more existing modes.
    Also accessible via the shortcut variable ``Del``.

    The deleted modes are traced out.
    After the deletion the state of the remaining subsystems may have to be described using a density operator.
    """

    ns = None

    def __or__(self, reg):
        reg = super().__or__(reg)
        pu.Program_current_context._delete_subsystems(reg)

    def _apply(self, reg, backend, **kwargs):
        backend.del_mode(reg)

    def __str__(self):
        # use the shorthand object
        return "Del"


def New(n=1):
    """Adds new subsystems to the quantum register.

    The new modes are prepared in the vacuum state.

    Must only be called in a :class:`Program` context.

    Args:
        n (int): number of subsystems to add
    Returns:
        tuple[RegRef]: tuple of the newly added subsystem references
    """
    if pu.Program_current_context is None:
        raise RuntimeError("New() can only be called inside a Program context.")
    # create RegRefs for the new modes
    refs = pu.Program_current_context._add_subsystems(n)
    # append the actual Operation to the Program
    pu.Program_current_context.append(_New_modes(n), refs)
    return refs


class _New_modes(MetaOperation):
    """Used internally for adding new modes to the system in a deferred way.

    This class cannot be used with the :meth:`__or__` syntax since it would be misleading.
    Indeed, users should *not* use this class directly, but rather the function :func:`New`.
    """

    ns = 0

    def __init__(self, n=1):
        """
        Args:
            n (int): number of modes to add
        """
        super().__init__()
        self.n = n  # int: store the number of new modes for the __str__ method

    def _apply(self, reg, backend, **kwargs):
        # pylint: disable=unused-variable
        inds = backend.add_mode(len(reg))

    def __str__(self):
        # use the shorthand object
        return "New({})".format(self.n)


class All(MetaOperation):
    """Metaoperation for applying a single-mode operation to every mode in the register.

    Args:
        op (Operation): single-mode operation to apply
    """

    def __init__(self, op):
        if op.ns != 1:
            raise ValueError("Not a one-subsystem operation.")
        super().__init__()
        self.op = op  #: Operation: one-subsystem operation to apply

    def __str__(self):
        return super().__str__() + "({})".format(str(self.op))

    def __or__(self, reg):
        # into a list of subsystems
        reg = _seq_to_list(reg)
        # convert into commands
        # make sure reg does not contain duplicates (we feed them to Program.append() one by one)
        pu.Program_current_context._test_regrefs(reg)
        for r in reg:
            pu.Program_current_context.append(self.op, [r])


# ====================================================================
# Decompositions
# ====================================================================


class Interferometer(Decomposition):
    r"""Apply a linear interferometer to the specified qumodes.

    This operation uses either the :ref:`rectangular decomposition <rectangular>`
    or triangular decomposition to decompose
    a linear interferometer into a sequence of beamsplitters and
    rotation gates.

    By specifying the keyword argument ``mesh``, the scheme used to implement the interferometer
    may be adjusted:

    * ``mesh='rectangular'`` (default): uses the scheme described in
      :cite:`clements2016`, resulting in a *rectangular* array of
      :math:`M(M-1)/2` beamsplitters:

      .. figure:: ../../_static/clements.png
          :align: center
          :width: 30%
          :target: javascript:void(0);

      Local phase shifts appear in the middle of the beamsplitter array.
      Use ``mesh='rectangular_phase_end`` to instead commute all local phase shifts
      to the end of the beamsplitter array.

      By default, the interferometers are decomposed into :class:`~.BSgate` operations.
      To instead decompose the interferometer using the :class:`~.ops.MZgate`,
      use ``mesh='rectangular_symmetric'``.

    * ``mesh='triangular'``: uses the scheme described in :cite:`reck1994`,
      resulting in a *triangular* array of :math:`M(M-1)/2` beamsplitters:

      .. figure:: ../../_static/reck.png
          :align: center
          :width: 30%
          :target: javascript:void(0);

      Local phase shifts appear at the end of the beamsplitter array.

    Args:
        U (array[complex]): an :math:`N\times N` unitary matrix
        mesh (str): the scheme used to implement the interferometer.
            Options include:

            - ``'rectangular'`` - rectangular mesh, with local phase shifts
              applied between interferometers

            - ``'rectangular_phase_end'`` - rectangular mesh, with local phase shifts
              placed after all interferometers

            - ``'rectangular_symmetric'`` - rectangular mesh, with local phase shifts
              placed after all interferometers, and all beamsplitters decomposed into
              pairs of symmetric beamsplitters and phase shifters

            - ``'triangular'`` - triangular mesh

        drop_identity (bool): If ``True``, decomposed gates with trivial parameters,
            such that they correspond to an identity operation, are removed.
        tol (float): the tolerance used when checking if the input matrix is unitary:
            :math:`|U-U^\dagger| <` tol
    """
    # pylint: disable=too-many-instance-attributes

    def __init__(self, U, mesh="rectangular", drop_identity=True, tol=1e-6):
        super().__init__([U])
        self.ns = U.shape[0]
        self.mesh = mesh
        self.tol = tol
        self.drop_identity = drop_identity

        allowed_meshes = {
            "rectangular",
            "rectangular_phase_end",
            "rectangular_symmetric",
            "triangular",
        }

        if mesh not in allowed_meshes:
            raise ValueError(
                "Unknown mesh '{}'. Mesh must be one of {}".format(mesh, allowed_meshes)
            )

        self.identity = np.allclose(U, np.identity(len(U)), atol=_decomposition_merge_tol, rtol=0)

    def _decompose(self, reg, **kwargs):
        mesh = kwargs.get("mesh", self.mesh)
        tol = kwargs.get("tol", self.tol)
        drop_identity = kwargs.get("drop_identity", self.drop_identity)

        cmds = []

        if not self.identity or not drop_identity:
            decomp_fn = getattr(dec, mesh)
            BS1, R, BS2 = decomp_fn(self.p[0], tol=tol)

            for n, m, theta, phi, _ in BS1:
                theta = theta if np.abs(theta) >= _decomposition_tol else 0
                phi = phi if np.abs(phi) >= _decomposition_tol else 0

                if "symmetric" in mesh:
                    # Mach-Zehnder interferometers
                    cmds.append(
                        Command(
                            MZgate(np.mod(theta, 2 * np.pi), np.mod(phi, 2 * np.pi)),
                            (reg[n], reg[m]),
                        )
                    )

                else:
                    # Clements style beamsplitters
                    if not (drop_identity and phi == 0):
                        cmds.append(Command(Rgate(phi), reg[n]))

                    if not (drop_identity and theta == 0):
                        cmds.append(Command(BSgate(theta, 0), (reg[n], reg[m])))

            for n, expphi in enumerate(R):
                # local phase shifts
                q = np.log(expphi).imag if np.abs(expphi - 1) >= _decomposition_tol else 0
                if not (drop_identity and q == 0):
                    cmds.append(Command(Rgate(q), reg[n]))

            if BS2 is not None:
                # Clements style beamsplitters

                for n, m, theta, phi, _ in reversed(BS2):
                    theta = theta if np.abs(theta) >= _decomposition_tol else 0
                    phi = phi if np.abs(phi) >= _decomposition_tol else 0

                    if not (drop_identity and theta == 0):
                        cmds.append(Command(BSgate(-theta, 0), (reg[n], reg[m])))
                    if not (drop_identity and phi == 0):
                        cmds.append(Command(Rgate(-phi), reg[n]))

        return cmds


class GraphEmbed(Decomposition):
    r"""Embed a graph into an interferometer setup.

    This operation uses the Takagi decomposition to decompose
    an adjacency matrix into a sequence of squeezers and beamsplitters and
    rotation gates.

    Args:
        A (array): an :math:`N\times N` complex or real symmetric matrix
        mean_photon_per_mode (float): guarantees that the mean photon number in the pure Gaussian state
            representing the graph satisfies  :math:`\frac{1}{N}\sum_{i=1}^N sinh(r_{i})^2 ==` :code:``mean_photon``
        make_traceless (boolean): Removes the trace of the input matrix, by performing the transformation
            :math:`\tilde{A} = A-\mathrm{tr}(A) \I/n`. This may reduce the amount of squeezing needed to encode
            the graph but will lead to different photon number statistics for events with more than
            one photon in any mode.
        tol (float): the tolerance used when checking if the input matrix is symmetric:
            :math:`|A-A^T| <` tol
    """

    def __init__(self, A, mean_photon_per_mode=1.0, make_traceless=False, tol=1e-6):
        super().__init__([A])
        self.ns = A.shape[0]

        if np.allclose(A, np.identity(len(A)), atol=_decomposition_merge_tol, rtol=0):
            self.identity = True
        else:
            self.identity = False
            self.sq, self.U = dec.graph_embed(
                A,
                mean_photon_per_mode=mean_photon_per_mode,
                make_traceless=make_traceless,
                atol=tol,
                rtol=0,
            )

    def _decompose(self, reg, **kwargs):
        cmds = []

        if not self.identity:
            for n, s in enumerate(self.sq):
                if np.abs(s) >= _decomposition_tol:
                    cmds.append(Command(Sgate(s), reg[n]))

            if not np.allclose(self.U, np.identity(len(self.U)), atol=_decomposition_tol, rtol=0):
                mesh = kwargs.get("mesh", "rectangular")
                cmds.append(Command(Interferometer(self.U, mesh=mesh), reg))

        return cmds


class BipartiteGraphEmbed(Decomposition):
    r"""Embed a bipartite graph into an interferometer setup.

    A bipartite graph is a graph that consists of two vertex sets :math:`U` and :math:`V`,
    such that every edge in the graph connects a vertex between :math:`U` and :math:`V`.
    That is, there are no edges between vertices in the same vertex set.

    The adjacency matrix of an :math:`N` vertex undirected bipartite graph
    is a :math:`N\times N` symmetric matrix of the form

    .. math:: A = \begin{bmatrix}0 & B \\ B^T & 0\end{bmatrix}

    where :math:`B` is a :math:N/2\times N/2` matrix representing the (weighted)
    edges between the vertex set.

    This operation decomposes an adjacency matrix into a sequence of two
    mode squeezers, beamsplitters, and rotation gates.

    Args:
        A (array): Either an :math:`N\times N` complex or real symmetric adjacency matrix
            :math:`A`, or an :math:`N/2\times N/2` complex or real matrix :math:`B`
            representing the edges between the vertex sets if ``edges=True``.
        mean_photon_per_mode (float): guarantees that the mean photon number in the pure Gaussian state
            representing the graph satisfies  :math:`\frac{1}{N}\sum_{i=1}^N sinh(r_{i})^2 ==` :code:``mean_photon``
        edges (bool): set to ``True`` if argument ``A`` represents the edges :math:`B`
            between the vertex sets rather than the full adjacency matrix
        drop_identity (bool): If ``True``, decomposed gates with trivial parameters,
            such that they correspond to an identity operation, are removed.
        tol (float): the tolerance used when checking if the input matrix is symmetric:
            :math:`|A-A^T| <` tol
    """

    def __init__(self, A, mean_photon_per_mode=1.0, edges=False, drop_identity=True, tol=1e-6):
        self._check_p0(A)
        self.mean_photon_per_mode = mean_photon_per_mode
        self.tol = tol
        self.identity = np.all(np.abs(A - np.identity(len(A))) < _decomposition_merge_tol)
        self.drop_identity = drop_identity

        if edges:
            self.ns = 2 * A.shape[0]
            B = A
        else:
            self.ns = A.shape[0]

            # check if A is a bipartite graph
            N = A.shape[0] // 2
            A00 = A[:N, :N]
            A11 = A[N:, N:]

            diag_zeros = np.allclose(A00, np.zeros_like(A00), atol=tol, rtol=0) and np.allclose(
                A11, np.zeros_like(A11), atol=tol, rtol=0
            )

            if (not diag_zeros) or (not np.allclose(A, A.T, atol=tol, rtol=0)):
                raise ValueError(
                    "Adjacency matrix {} does not represent a bipartite graph".format(A)
                )

            B = A[:N, N:]

        super().__init__([B])

    def _decompose(self, reg, **kwargs):
        mean_photon_per_mode = kwargs.get("mean_photon_per_mode", self.mean_photon_per_mode)
        tol = kwargs.get("tol", self.tol)
        mesh = kwargs.get("mesh", "rectangular")
        drop_identity = kwargs.get("drop_identity", self.drop_identity)

        cmds = []

        B = self.p[0]
        N = len(B)

        sq, U, V = dec.bipartite_graph_embed(
            B, mean_photon_per_mode=mean_photon_per_mode, atol=tol, rtol=0
        )

        if not self.identity or not drop_identity:
            for m, s in enumerate(sq):
                s = s if np.abs(s) >= _decomposition_tol else 0

                if not (drop_identity and s == 0):
                    cmds.append(Command(S2gate(-s), (reg[m], reg[m + N])))

            for X, _reg in ((U, reg[:N]), (V, reg[N:])):

                if np.allclose(X, np.identity(len(X)), atol=_decomposition_tol, rtol=0):
                    X = np.identity(len(X))

                if not (drop_identity and np.all(X == np.identity(len(X)))):
                    cmds.append(
                        Command(
                            Interferometer(X, mesh=mesh, drop_identity=drop_identity, tol=tol), _reg
                        )
                    )

        return cmds


class GaussianTransform(Decomposition):
    r"""Apply a Gaussian symplectic transformation to the specified qumodes.

    This operation uses the :ref:`Bloch-Messiah decomposition <bloch_messiah>`
    to decompose a symplectic matrix :math:`S`:

    .. math:: S = O_1 R O_2

    where :math:`O_1` and :math:`O_2` are two orthogonal symplectic matrices (and thus passive
    Gaussian transformations), and :math:`R`
    is a squeezing transformation in the phase space (:math:`R=\text{diag}(e^{-z},e^z)`).

    The symplectic matrix describing the Gaussian transformation on :math:`N` modes must satisfy

    .. math:: S\Omega S^T = \Omega, ~~\Omega = \begin{bmatrix}0&I\\-I&0\end{bmatrix}

    where :math:`I` is the :math:`N\times N` identity matrix, and :math:`0` is the zero matrix.

    The two orthogonal symplectic unitaries describing the interferometers are then further
    decomposed via the :class:`~.Interferometer` operator and the
    :ref:`Rectangular decomposition <rectangular>`:

    .. math:: U_i = X_i + iY_i

    where

    .. math:: O_i = \begin{bmatrix}X&-Y\\Y&X\end{bmatrix}

    Args:
        S (array[float]): a :math:`2N\times 2N` symplectic matrix describing the Gaussian transformation.
        vacuum (bool): set to True if acting on a vacuum state. In this case, :math:`O_2 V O_2^T = I`,
            and the unitary associated with orthogonal symplectic :math:`O_2` will be ignored.
        tol (float): the tolerance used when checking if the matrix is symplectic:
            :math:`|S^T\Omega S-\Omega| \leq` tol
    """

    def __init__(self, S, vacuum=False, tol=1e-10):
        super().__init__([S])
        self.ns = S.shape[0] // 2
        self.vacuum = (
            vacuum  #: bool: if True, ignore the first unitary matrix when applying the gate
        )
        N = self.ns  # shorthand

        # check if input symplectic is passive (orthogonal)
        diffn = np.linalg.norm(S @ S.T - np.identity(2 * N))
        self.active = (
            np.abs(diffn) > _decomposition_tol
        )  #: bool: S is an active symplectic transformation

        if not self.active:
            # The transformation is passive, do Clements
            X1 = S[:N, :N]
            P1 = S[N:, :N]
            self.U1 = X1 + 1j * P1
        else:
            # transformation is active, do Bloch-Messiah
            O1, smat, O2 = dec.bloch_messiah(S, tol=tol)
            X1 = O1[:N, :N]
            P1 = O1[N:, :N]
            X2 = O2[:N, :N]
            P2 = O2[N:, :N]

            self.U1 = X1 + 1j * P1  #: array[complex]: unitary matrix corresponding to O_1
            self.U2 = X2 + 1j * P2  #: array[complex]: unitary matrix corresponding to O_2
            self.Sq = np.diagonal(smat)[
                :N
            ]  #: array[complex]: diagonal vector of the squeezing matrix R

    def _decompose(self, reg, **kwargs):
        cmds = []
        mesh = kwargs.get("mesh", "rectangular")

        if self.active:
            if not self.vacuum:
                cmds = [Command(Interferometer(self.U2), reg)]

            for n, expr in enumerate(self.Sq):
                if np.abs(expr - 1) >= _decomposition_tol:
                    r = np.abs(np.log(expr))
                    phi = np.angle(np.log(expr))
                    cmds.append(Command(Sgate(-r, phi), reg[n]))

            cmds.append(Command(Interferometer(self.U1, mesh=mesh), reg))
        else:
            if not self.vacuum:
                cmds = [Command(Interferometer(self.U1, mesh=mesh), reg)]

        return cmds


class Gaussian(Preparation, Decomposition):
    r"""Prepare the specified modes in a Gaussian state.

    This operation uses the :ref:`Williamson decomposition <williamson>` to prepare
    quantum modes into a given Gaussian state, specified by a
    vector of means and a covariance matrix.

    The Williamson decomposition decomposes the Gaussian state into a Gaussian
    transformation (represented by a symplectic matrix) acting on :class:`~.Thermal`
    states. The Gaussian transformation is then further decomposed into an array
    of beamsplitters and local squeezing and rotation gates, by way of the
    :class:`~.GaussianTransform` and :class:`~.Interferometer` decompositions.

    Alternatively, the decomposition can be explicitly turned off, and the
    backend can be explicitly prepared in the Gaussian state provided. This is
    **only** supported by backends using the Gaussian representation.

    Args:
        V (array[float]): an :math:`2N\times 2N` (real and positive definite) covariance matrix
        r (array[float] or None): Length :math:`2N` vector of means, of the
            form :math:`(\x_0,\dots,\x_{N-1},\p_0,\dots,\p_{N-1})`.
            If None, it is assumed that :math:`r=0`.
        decomp (bool): Should the operation be decomposed into a sequence of elementary gates?
            If False, the state preparation is performed directly via the backend API.
        tol (float): the tolerance used when checking if the matrix is symmetric: :math:`|V-V^T| \leq` tol
    """
    # pylint: disable=too-many-instance-attributes
    ns = None

    def __init__(self, V, r=None, decomp=True, tol=1e-6):
        self._check_p0(V)
        # internally we eliminate hbar from the covariance matrix V (or equivalently set hbar=2), but not from the means vector r
        V = V / (sf.hbar / 2)
        self.ns = V.shape[0] // 2

        if r is None:
            r = np.zeros(2 * self.ns)
        r = np.asarray(r)

        if len(r) != V.shape[0]:
            raise ValueError("Vector of means must have the same length as the covariance matrix.")

        super().__init__([V, r], decomp=decomp)  # V is hbar-independent, r is not

        self.x_disp = r[: self.ns]
        self.p_disp = r[self.ns :]

        # needed only if decomposed
        th, self.S = dec.williamson(V, tol=tol)
        self.pure = np.abs(np.linalg.det(V) - 1.0) < tol
        self.nbar = 0.5 * (np.diag(th)[: self.ns] - 1.0)

    def _apply(self, reg, backend, **kwargs):
        p = par_evaluate(self.p)
        s = np.sqrt(sf.hbar / 2)  # scaling factor, since the backend API call is hbar-independent
        backend.prepare_gaussian_state(p[1] / s, p[0], reg)

    def _decompose(self, reg, **kwargs):
        # pylint: disable=too-many-branches
        cmds = []

        V = self.p[0]
        D = np.diag(V)
        is_diag = np.all(V == np.diag(D))

        BD = changebasis(self.ns) @ V @ changebasis(self.ns).T
        BD_modes = [BD[i * 2 : (i + 1) * 2, i * 2 : (i + 1) * 2] for i in range(BD.shape[0] // 2)]
        is_block_diag = (not is_diag) and np.all(BD == block_diag(*BD_modes))

        if self.pure and is_diag:
            # covariance matrix consists of x/p quadrature squeezed state
            for n, expr in enumerate(D[: self.ns]):
                if np.abs(expr - 1) >= _decomposition_tol:
                    r = np.abs(np.log(expr) / 2)
                    cmds.append(Command(Squeezed(r, 0), reg[n]))
                else:
                    cmds.append(Command(Vac, reg[n]))

        elif self.pure and is_block_diag:
            # covariance matrix consists of rotated squeezed states
            for n, v in enumerate(BD_modes):
                if not np.all(v - np.identity(2) < _decomposition_tol):
                    r = np.abs(np.arccosh(np.sum(np.diag(v)) / 2)) / 2
                    phi = np.arctan(2 * v[0, 1] / np.sum(np.diag(v) * [1, -1]))
                    cmds.append(Command(Squeezed(r, phi), reg[n]))
                else:
                    cmds.append(Command(Vac, reg[n]))

        elif not self.pure and is_diag and np.all(D[: self.ns] == D[self.ns :]):
            # covariance matrix consists of thermal states
            for n, nbar in enumerate(0.5 * (D[: self.ns] - 1.0)):
                if nbar >= _decomposition_tol:
                    cmds.append(Command(Thermal(nbar), reg[n]))
                else:
                    cmds.append(Command(Vac, reg[n]))

        else:
            if not self.pure:
                # mixed state, must initialise thermal states
                for n, nbar in enumerate(self.nbar):
                    if np.abs(nbar) >= _decomposition_tol:
                        cmds.append(Command(Thermal(nbar), reg[n]))
                    else:
                        cmds.append(Command(Vac, reg[n]))

            else:
                for r in reg:
                    cmds.append(Command(Vac, r))

            cmds.append(Command(GaussianTransform(self.S, vacuum=self.pure), reg))

        cmds += [Command(Xgate(u), reg[n]) for n, u in enumerate(self.x_disp) if u != 0]
        cmds += [Command(Zgate(u), reg[n]) for n, u in enumerate(self.p_disp) if u != 0]

        return cmds


# =======================================================================
# Shorthands, e.g. pre-constructed singleton-like objects

Del = _Delete()
Vac = Vacuum()
MeasureX = MeasureHomodyne(0)
MeasureP = MeasureHomodyne(np.pi / 2)
MeasureHD = MeasureHeterodyne()

Fourier = Fouriergate()

shorthands = ["New", "Del", "Vac", "MeasureX", "MeasureP", "MeasureHD", "Fourier", "All"]

# =======================================================================
# here we list different classes of operations for unit testing purposes

zero_args_gates = (Fouriergate,)
one_args_gates = (Xgate, Zgate, Rgate, Pgate, Vgate, Kgate, CXgate, CZgate, CKgate)
two_args_gates = (Dgate, Sgate, BSgate, MZgate, S2gate)
gates = zero_args_gates + one_args_gates + two_args_gates

channels = (LossChannel, ThermalLossChannel)

simple_state_preparations = (
    Vacuum,
    Coherent,
    Squeezed,
    DisplacedSqueezed,
    Fock,
    Catstate,
    Thermal,
)  # have __init__ methods with default arguments
state_preparations = simple_state_preparations + (Ket, DensityMatrix)

measurements = (MeasureFock, MeasureHomodyne, MeasureHeterodyne, MeasureThreshold)

decompositions = (Interferometer, BipartiteGraphEmbed, GraphEmbed, GaussianTransform, Gaussian)

# =======================================================================
# exported symbols

__all__ = [
    cls.__name__ for cls in gates + channels + state_preparations + measurements + decompositions
] + shorthands



# Copyright 2019 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
r"""
Functionality for generating GBS samples using classical simulators.

.. seealso::

    :ref:`apps-sample-tutorial`

Generating samples
------------------

An :math:`M` mode GBS device can be programmed by specifying an :math:`(M \times M)`-dimensional
symmetric matrix :math:`A` :cite:`bradler2018gaussian`. Running this device results in samples
that carry relevant information about the encoded matrix :math:`A`. When sampling, one must also
specify the mean number of photons in the device, the form of detection used at the output:
threshold detection or photon-number-resolving (PNR) detection, as well as the amount of loss.

The :func:`sample` function provides a simulation of sampling from GBS.

Here, each output sample is an :math:`M`-dimensional list. If threshold detection is used
(``threshold = True``), each element of a sample is either a zero (denoting no photons detected)
or a one (denoting one or more photons detected), conventionally called a "click".
If photon-number resolving (PNR) detection is used (``threshold = False``) then elements of a
sample are non-negative integers counting the number of photons detected in each mode. The ``loss``
parameter allows for simulation of photon loss in the device, which is the most common form of
noise in quantum photonics. Here, ``loss = 0`` describes ideal loss-free GBS, while generally
``0 <= loss <= 1`` describes the proportion of photons lost while passing through the device.

Samples can be postselected based upon their total number of photons or clicks and the ``numpy``
random seed used to generate samples can be fixed.

The :func:`vibronic` function allows users to sample Gaussian states for computing molecular
vibronic spectra.

Generating subgraphs
--------------------

There are two forms to represent a sample from a GBS device:

1. In the form returned by :func:`sample`, each sample is a length :math:`M` list of counts in
   each mode, e.g., the sample ``[2, 1, 2, 0, 1]`` denotes two photons in mode 0,
   1 photon in mode 1, and so forth.

2. The alternative representation is a list of modes where photons or clicks were observed, e.g.,
   the above sample can be written alternatively as ``[0, 0, 1, 2, 2, 4]``.

Converting from the former representation to the latter can be achieved with:

.. autosummary::
    modes_from_counts

Graphs can be encoded into GBS through their adjacency matrix. Resultant samples
can then be used to pick out nodes from the graph to form a subgraph. If threshold detection is
used, any nodes that click are selected as part of the subgraph. For example, if a sample is
``[1, 1, 1, 0, 1]`` then the corresponding subgraph has nodes ``[0, 1, 2, 4]``. This can be found
using:

>>> modes_from_counts([1, 1, 1, 0, 1])
[0, 1, 2, 4]

A collection of GBS samples from a graph, given by :func:`sample` in the first form, can be
converted to subgraphs using:

.. autosummary::
    to_subgraphs

A typical workflow would be:

>>> g = nx.erdos_renyi_graph(5, 0.7)
>>> a = nx.to_numpy_array(g)
>>> s = sample(a, 3, 4)
>>> s = to_subgraphs(s, g)
[[0, 2], [0, 1, 2, 4], [1, 2, 3, 4], [1]]

The subgraphs sampled from GBS are likely to be dense :cite:`arrazola2018using`, motivating their
use within heuristics for problems such as maximum clique (see :mod:`~.apps.clique`).

Including node weights
----------------------

Some graphs are composed of nodes with weights. These weights can correspond to relevant
information in an optimization problem and it is desirable to encode them into GBS along with the
graph's adjacency matrix. One canonical approach to doing this is to use the :math:`WAW` encoding
:cite:`banchi2019molecular`, which rescales the adjacency matrix according to:

.. math::

    A \rightarrow WAW,

with :math:`W=w_{i}\delta_{ij}` the diagonal matrix formed by the weighted nodes :math:`w_{i}`. The
rescaled adjacency matrix can be passed to :func:`sample`, resulting in a distribution that
increases the probability of observing a node proportionally to its weight.
"""
import warnings
from typing import Optional, Union

import networkx as nx
import numpy as np

import strawberryfields as sf


def sample(
    A: np.ndarray, n_mean: float, n_samples: int = 1, threshold: bool = True, loss: float = 0.0
) -> list:
    r"""Generate simulated samples from GBS encoded with a symmetric matrix :math:`A`.

    **Example usage:**

    >>> g = nx.erdos_renyi_graph(5, 0.7)
    >>> a = nx.to_numpy_array(g)
    >>> sample(a, 3, 4)
    [[1, 1, 1, 1, 1], [1, 1, 0, 1, 1], [0, 0, 0, 0, 0], [1, 0, 0, 0, 1]]

    Args:
        A (array): the symmetric matrix to sample from
        n_mean (float): mean photon number
        n_samples (int): number of samples
        threshold (bool): perform GBS with threshold detectors if ``True`` or photon-number
            resolving detectors if ``False``
        loss (float): fraction of generated photons that are lost while passing through device.
            Parameter should range from ``loss=0`` (ideal noise-free GBS) to ``loss=1``.

    Returns:
        list[list[int]]: a list of samples from GBS with respect to the input symmetric matrix
    """
    if not np.allclose(A, A.T):
        raise ValueError("Input must be a NumPy array corresponding to a symmetric matrix")
    if n_samples < 1:
        raise ValueError("Number of samples must be at least one")
    if n_mean < 0:
        raise ValueError("Mean photon number must be non-negative")
    if not 0 <= loss <= 1:
        raise ValueError("Loss parameter must take a value between zero and one")

    nodes = len(A)

    p = sf.Program(nodes)
    eng = sf.LocalEngine(backend="gaussian")

    mean_photon_per_mode = n_mean / float(nodes)

    # pylint: disable=expression-not-assigned,pointless-statement
    with p.context as q:
        sf.ops.GraphEmbed(A, mean_photon_per_mode=mean_photon_per_mode) | q

        if loss:
            for _q in q:
                sf.ops.LossChannel(1 - loss) | _q

        if threshold:
            sf.ops.MeasureThreshold() | q
        else:
            sf.ops.MeasureFock() | q

    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", category=UserWarning, message="Cannot simulate non-")
        s = eng.run(p, shots=n_samples).samples
    return s.tolist()


def postselect(samples: list, min_count: int, max_count: int) -> list:
    """Postselect samples by imposing a minimum and maximum number of photons or clicks.

    **Example usage:**

    >>> s = [[1, 1, 1, 1, 1], [1, 1, 0, 1, 1], [0, 0, 0, 0, 0], [1, 0, 0, 0, 1]]
    >>> postselect(s, 2, 4)
    [[1, 1, 0, 1, 1], [1, 0, 0, 0, 1]]

    Args:
        samples (list[list[int]]): a list of samples
        min_count (int): minimum number of photons or clicks for a sample to be included
        max_count (int): maximum number of photons or clicks for a sample to be included

    Returns:
        list[list[int]]: the postselected samples
    """
    return [s for s in samples if min_count <= sum(s) <= max_count]


def seed(value: Optional[int]) -> None:
    """Seed for random number generators.

    Wrapper function for `numpy.random.seed <https://docs.scipy.org/doc/numpy//reference/generated
    /numpy.random.seed.html>`_ to seed all NumPy-based random number generators. This allows for
    repeatable sampling.

    **Example usage:**

    >>> g = nx.erdos_renyi_graph(5, 0.7)
    >>> a = nx.to_numpy_array(g)
    >>> seed(1967)
    >>> sample(a, 3, 4)
    [[1, 1, 1, 1, 1], [1, 1, 1, 1, 1], [1, 0, 1, 0, 1], [0, 0, 0, 0, 0]]
    >>> seed(1967)
    >>> sample(a, 3, 4)
    [[1, 1, 1, 1, 1], [1, 1, 1, 1, 1], [1, 0, 1, 0, 1], [0, 0, 0, 0, 0]]

    Args:
        value (int): random seed
    """
    np.random.seed(value)


def modes_from_counts(s: list) -> list:
    r"""Convert a list of counts to a list of modes where photons or clicks were observed.

    Consider an :math:`M`-mode sample :math:`s=\{s_{0},s_{1},\ldots,s_{M-1}\}` of counts in each
    mode, i.e., so that mode :math:`i` has :math:`s_{i}` photons/clicks. If :math:`k = \sum_{
    i=0}^{M-1} s_{i}` is the total number of photons or clicks, this function returns a list of
    modes :math:`m=\{m_{1},m_{2},\ldots, m_{k}\}` where photons or clicks were observed, i.e.,
    so that photon/click :math:`i` is in mode :math:`m_{i}`. Since there are typically fewer
    photons than modes, the latter form often gives a shorter representation.

    **Example usage:**

    >>> modes_from_counts([0, 1, 0, 1, 2, 0])
    [1, 3, 4, 4]

    Args:
       s (list[int]): a sample of counts

    Returns:
        list[int]: a list of modes where photons or clicks were observed, sorted in
        non-decreasing order
    """
    modes = []
    for i, c in enumerate(s):
        modes += [i] * int(c)
    return sorted(modes)


def to_subgraphs(samples: list, graph: nx.Graph) -> list:
    """Converts samples to their subgraph representation.

    Input samples are a list of counts that are processed into subgraphs by selecting the nodes
    where a click occurred.

    **Example usage:**

    >>> g = nx.erdos_renyi_graph(5, 0.7)
    >>> s = [[1, 1, 1, 1, 1], [1, 1, 0, 1, 1], [0, 0, 0, 0, 0], [1, 0, 0, 0, 1]]
    >>> to_subgraphs(s, g)
    [[0, 1, 2, 3, 4], [0, 1, 3, 4], [], [0, 4]]

    Args:
        samples (list[list[int]]): a list of samples, each sample being a list of counts
        graph (nx.Graph): the input graph

    Returns:
        list[list[int]]: a list of subgraphs, where each subgraph is represented by a list of its
        nodes
    """

    graph_nodes = list(graph.nodes)
    node_number = len(graph_nodes)

    subgraph_samples = [list(set(modes_from_counts(s))) for s in samples]

    if graph_nodes != list(range(node_number)):
        return [sorted([graph_nodes[i] for i in s]) for s in subgraph_samples]

    return subgraph_samples


def vibronic(
    t: np.ndarray,
    U1: np.ndarray,
    r: np.ndarray,
    U2: np.ndarray,
    alpha: np.ndarray,
    n_samples: int,
    loss: float = 0.0,
) -> list:
    """Generate samples for computing vibronic spectra. The following gates are applied to input
    vacuum states:

    1. Two-mode squeezing on all :math:`2N` modes with parameters ``t``
    2. Interferometer ``U1`` on the first :math:`N` modes
    3. Squeezing on the first :math:`N` modes with parameters ``r``
    4. Interferometer ``U2`` on the first :math:`N` modes
    5. Displacement on the first :math:`N` modes with parameters ``alpha``

    A sample is generated by measuring the number of photons in each of the :math:`2N` modes. In
    the special case that all of the two-mode squeezing parameters ``t`` are zero, only :math:`N`
    modes are considered, which speeds up calculations.

    **Example usage:**

    >>> formic = data.Formic()
    >>> w = formic.w
    >>> wp = formic.wp
    >>> Ud = formic.Ud
    >>> delta = formic.delta
    >>> T = 0
    >>> t, U1, r, U2, alpha = vibronic.gbs_params(w, wp, Ud, delta, T)
    >>> sample.vibronic(t, U1, r, U2, alpha, 2, 0.0)
    [[0, 0, 2, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0],
     [0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]]

    Args:
        t (array): two-mode squeezing parameters
        U1 (array): unitary matrix for the first interferometer
        r (array): squeezing parameters
        U2 (array): unitary matrix for the second interferometer
        alpha (array): displacement parameters
        n_samples (int): number of samples to be generated
        loss (float): loss parameter denoting the fraction of generated photons that are lost


    Returns:
        list[list[int]]: a list of samples from GBS
    """
    if n_samples < 1:
        raise ValueError("Number of samples must be at least one")
    if not 0 <= loss <= 1:
        raise ValueError("Loss parameter must take a value between zero and one")

    n_modes = len(t)

    eng = sf.LocalEngine(backend="gaussian")

    if np.any(t != 0):
        gbs = sf.Program(n_modes * 2)
    else:
        gbs = sf.Program(n_modes)

    # pylint: disable=expression-not-assigned,pointless-statement
    with gbs.context as q:

        if np.any(t != 0):
            for i in range(n_modes):
                sf.ops.S2gate(t[i]) | (q[i], q[i + n_modes])

        sf.ops.Interferometer(U1) | q[:n_modes]

        for i in range(n_modes):
            sf.ops.Sgate(r[i]) | q[i]

        sf.ops.Interferometer(U2) | q[:n_modes]

        for i in range(n_modes):
            sf.ops.Dgate(alpha[i]) | q[i]

        if loss:
            for _q in q:
                sf.ops.LossChannel(1 - loss) | _q

        sf.ops.MeasureFock() | q

    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", category=UserWarning, message="Cannot simulate non-")

        s = eng.run(gbs, shots=n_samples).samples

    s = np.array(s).tolist()  # convert all generated samples to list

    if np.any(t == 0):
        s = np.pad(s, ((0, 0), (0, n_modes))).tolist()
    return s


def waw_matrix(A: np.ndarray, w: Union[np.ndarray, list]) -> np.ndarray:
    r"""Rescale adjacency matrix to account for node weights.

    Given a graph with adjacency matrix :math:`A` and a vector :math:`w` of weighted nodes,
    this function rescales the adjacency matrix according to :cite:`banchi2019molecular`:

    .. math::

        A \rightarrow WAW,

    with :math:`W=w_{i}\delta_{ij}` the diagonal matrix formed by the weighted nodes :math:`w_{i}`.
    The rescaled adjacency matrix can be passed to :func:`sample`, resulting in a distribution that
    increases the probability of observing a node proportionally to its weight.

    **Example usage:**

    >>> g = nx.erdos_renyi_graph(5, 0.7)
    >>> a = nx.to_numpy_array(g)
    >>> a
    array([[0., 1., 1., 1., 1.],
           [1., 0., 1., 1., 1.],
           [1., 1., 0., 1., 0.],
           [1., 1., 1., 0., 0.],
           [1., 1., 0., 0., 0.]])
    >>> w = [10, 1, 0, 1, 1]
    >>> a = waw_matrix(a, w)
    >>> a
    array([[ 0., 10.,  0., 10., 10.],
           [10.,  0.,  0.,  1.,  1.],
           [ 0.,  0.,  0.,  0.,  0.],
           [10.,  1.,  0.,  0.,  0.],
           [10.,  1.,  0.,  0.,  0.]])
    >>> sample(a, 3, 4)
    [[1, 0, 0, 1, 1], [1, 1, 0, 0, 0], [1, 1, 0, 1, 1], [1, 0, 0, 1, 0]]

    Args:
        A (array): adjacency matrix to rescale
        w (array or list): vector of real node weights

    Returns:
        array: matrix rescaled according to the WAW encoding
    """
    if not np.allclose(A, A.T):
        raise ValueError("Input must be a NumPy array corresponding to a symmetric matrix")

    if isinstance(w, list):
        w = np.array(w)

    return (w * A).T * w



# Copyright 2019-2020 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
This module provides a class that represents the result of a quantum computation.
"""
import numpy as np


class Result:
    """Result of a quantum computation.

    Represents the results of the execution of a quantum program on a local or
    remote backend.

    The returned :class:`~Result` object provides several useful properties
    for accessing the results of your program execution:

    * ``results.state``: The quantum state object contains details and methods
      for manipulation of the final circuit state. Not available for remote
      backends. See :doc:`/introduction/states` for more information regarding available
      state methods.

    * ``results.samples``: Measurement samples from any measurements performed.

    **Example:**

    The following example runs an existing Strawberry Fields
    quantum :class:`~.Program` on the Gaussian backend to get
    a ``Result`` object.

    Using this ``Result`` object, the measurement samples
    can be returned, as well as quantum state information.

    >>> eng = sf.Engine("gaussian")
    >>> results = eng.run(prog)
    >>> print(results)
    Result: 3 subsystems
        state: <GaussianState: num_modes=3, pure=True, hbar=2>
        samples: [[0, 0, 0]]
    >>> results.samples
    np.array([[0, 0, 0]])
    >>> results.state.is_pure()
    True

    .. note::

        Only local simulators will return a state object. Remote
        simulators and hardware backends will return
        measurement samples  (:attr:`Result.samples`),
        but the return value of ``Result.state`` will be ``None``.
    """

    def __init__(self, samples, is_stateful=True):
        self._state = None
        self._is_stateful = is_stateful

        # samples arrives as either a 2-D array (for shots > 1) or a 1-D array (for shots = 1)
        # the latter needs to be converted to a 2-D array with shape (shots, modes)
        if np.ndim(samples) == 1:
            samples = np.array([samples])
        self._samples = samples

    @property
    def samples(self):
        """Measurement samples.

        Returned measurement samples will have shape ``(shots, modes)``.

        Returns:
            array[array[float, int]]: measurement samples returned from
            program execution
        """
        return self._samples

    @property
    def state(self):
        """The quantum state object.

        The quantum state object contains details and methods
        for manipulation of the final circuit state.

        See :doc:`/introduction/states` for more information regarding available
        state methods.

        .. note::

            Only local simulators will return a state object. Remote
            simulators and hardware backends will return
            measurement samples (:attr:`Result.samples`),
            but the return value of ``Result.state`` will be ``None``.

        Returns:
            BaseState: quantum state returned from program execution
        """
        if not self._is_stateful:
            raise AttributeError("The state is undefined for a stateless computation.")
        return self._state

    def __str__(self):
        """String representation."""
        return "Result: {} subsystems, state: {}\n samples: {}".format(
            len(self.samples), self.state, self.samples
        )



#!/usr/bin/env python3
import strawberryfields as sf
from strawberryfields.ops import *
import tensorflow as tf

raise NotImplementedError('This example is unavailable until SF supports TensorFlow 2.0.')


# initialize engine and program objects
eng = sf.Engine(backend="tf", backend_options={"cutoff_dim": 7})
circuit = sf.Program(1)

tf_alpha = tf.Variable(0.1)
tf_phi = tf.Variable(0.1)

alpha, phi = circuit.params('alpha', 'phi')
with circuit.context as q:
    Dgate(alpha, phi) | q[0]
results = eng.run(circuit, args={'alpha': tf_alpha, 'phi': tf_phi}, eval=False})

# loss is probability for the coherent state |alpha=1>
prob = results.state.fidelity_coherent([1.0])
loss = -prob  # negative sign to maximize prob

# Set up optimization
optimizer = tf.train.GradientDescentOptimizer(learning_rate=0.1)
minimize_op = optimizer.minimize(loss)

# Create Tensorflow Session and initialize variables
sess = tf.Session()
sess.run(tf.global_variables_initializer())

# Carry out optimization
for step in range(50):
    prob_val, _ = sess.run([prob, minimize_op])
    print("Value at step {}: {}".format(step, prob_val))



# Copyright 2020 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
This module provides classes for interfacing with program execution jobs on a remote backend.
"""
import enum
import logging

from .result import Result

log = logging.getLogger(__name__)


class InvalidJobOperationError(Exception):
    """Raised when an invalid operation is performed on a job."""


class JobStatus(enum.Enum):
    """Represents the status of a remote job.

    This class maps a set of job statuses to the string representations returned by the
    remote platform.
    """

    OPEN = "open"
    QUEUED = "queued"
    CANCELLED = "cancelled"
    COMPLETED = "complete"
    FAILED = "failed"

    @property
    def is_final(self) -> bool:
        """Checks if this status represents a final and immutable state.

        This method is primarily used to determine if an operation is valid for a given
        status.

        Returns:
            bool
        """
        return self in (JobStatus.CANCELLED, JobStatus.COMPLETED, JobStatus.FAILED)

    def __repr__(self) -> str:
        return "<JobStatus: {}>".format(self.value)

    def __str__(self) -> str:
        return self.__repr__()


class Job:
    """Represents a remote job that can be queried for its status or result.

    This object should typically not be instantiated directly, but returned by an
    :class:`strawberryfields.RemoteEngine` or :class:`strawberryfields.api.Connection`
    when a job is run.

    Args:
        id_ (str): the job ID
        status (strawberryfields.api.JobStatus): the job status
        connection (strawberryfields.api.Connection): the connection over which the
            job is managed
    """

    def __init__(self, id_: str, status: JobStatus, connection):
        self._id = id_
        self._status = status
        self._connection = connection
        self._result = None

    @property
    def id(self) -> str:
        """The job ID.

        Returns:
            str
        """
        return self._id

    @property
    def status(self) -> str:
        """Returns the current job status, with possible values including ``"open"``,
        ``"queued"``, ``"cancelled"``, ``"complete"``, and ``"failed"``.

        Returns:
            str
        """
        return self._status.value

    @property
    def result(self) -> Result:
        """Returns the :class:`~.Result` of a completed job.

        This is only defined for completed jobs, and raises an exception for any other
        status.

        Returns:
            strawberryfields.api.Result
        """
        if self._status != JobStatus.COMPLETED:
            raise AttributeError(
                "The result is undefined for jobs that are not completed "
                "(current status: {})".format(self._status.value)
            )
        return self._result

    def refresh(self):
        """Refreshes the status of an open or queued job,
        along with the job result if the job is newly completed.
        """
        if self._status.is_final:
            log.warning("A %s job cannot be refreshed", self._status.value)
            return
        self._status = JobStatus(self._connection.get_job_status(self.id))
        if self._status == JobStatus.COMPLETED:
            self._result = self._connection.get_job_result(self.id)

    def cancel(self):
        """Cancels an open or queued job.

        Only an open or queued job can be cancelled; an exception is raised otherwise.
        """
        if self._status.is_final:
            raise InvalidJobOperationError(
                "A {} job cannot be cancelled".format(self._status.value)
            )
        self._connection.cancel_job(self.id)

    def __repr__(self):
        return "<{}: id={}, status={}>".format(self.__class__.__name__, self.id, self._status.value)

    def __str__(self):
        return self.__repr__()



# Copyright 2019 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
r"""
Tools for building kernel matrices and generating point processes using GBS.

.. seealso::

    :ref:`apps-points-tutorial`

Point processes
---------------

A point process is a mechanism that generates random point patterns among a set of possible
outcomes.
Point processes are statistical models that can replicate the stochastic
properties of natural phenomena, or be used as subroutines in statistical and machine learning
algorithms.

Several point processes rely on matrix functions to assign probabilities to different point
patterns. As shown in Ref. :cite:`jahangiri2019point`, GBS naturally gives rise to a *hafnian*
point process that employs the `hafnian
<https://the-walrus.readthedocs.io/en/latest/hafnian.html>`_ as the underlying matrix function.
This point process has the central property of generating clustered data points with
high probability. In this setting, a GBS device is programmed according to a *kernel* matrix that
encodes information about the similarity between points. When this kernel matrix is positive
semidefinite, it is possible to use GBS to implement a *permanental* point process and employ
fast classical algorithms to sample from the resulting distribution.

One choice of kernel matrix is the radial basis function (RBF) kernel whose elements are computed
as:

.. math::
    K_{i,j} = e^{-\|\bf{r}_i-\bf{r}_j\|^2/(2\sigma^2)},

where :math:`\bf{r}_i` are the coordinates of point :math:`i`, :math:`\sigma` is a kernel
parameter, and :math:`\|\cdot\|` denotes a choice of norm. The RBF kernel is positive
semidefinite when the Euclidean norm is used, as is the case for the provided :func:`rbf_kernel`
function.
"""

import numpy as np
import scipy
from thewalrus.csamples import generate_thermal_samples, rescale_adjacency_matrix_thermal


def rbf_kernel(R: np.ndarray, sigma: float) -> np.ndarray:
    r"""Calculate the RBF kernel matrix from a set of input points.

    The kernel parameter :math:`\sigma` is used to define the kernel scale. Points that are much
    further than :math:`\sigma` from each other lead to small entries of the kernel
    matrix, whereas points much closer than :math:`\sigma` generate large entries.

    The Euclidean norm is used to measure distance in this function, resulting in a
    positive-semidefinite kernel.

    **Example usage:**

    >>> R = np.array([[0, 1], [1, 0], [0, 0], [1, 1]])
    >>> rbf_kernel (R, 1.0)
    array([[1., 0.36787944, 0.60653066, 0.60653066],
           [0.36787944, 1., 0.60653066, 0.60653066],
           [0.60653066, 0.60653066, 1., 0.36787944],
           [0.60653066, 0.60653066, 0.36787944, 1.,]])

    Args:
        R (array): Coordinate matrix. Rows of this array are the coordinates of the points.
        sigma (float): kernel parameter

    Returns:
        K (array): the RBF kernel matrix
    """
    return np.exp(-((scipy.spatial.distance.cdist(R, R)) ** 2) / 2 / sigma ** 2)


def sample(K: np.ndarray, n_mean: float, n_samples: int) -> list:
    """Sample subsets of points using the permanental point process.

    Points can be encoded through a radial basis function kernel, provided in :func:`rbf_kernel`.
    Subsets of points are sampled with probabilities that are proportional to the permanent of the
    submatrix of the kernel selected by those points.

    This permanental point process is likely to sample points that are clustered together
    :cite:`jahangiri2019point`. It can be realized using a variant of Gaussian boson sampling
    with thermal states as input.

    **Example usage:**

    >>> K = np.array([[1., 0.36787944, 0.60653066, 0.60653066],
    >>>               [0.36787944, 1., 0.60653066, 0.60653066],
    >>>               [0.60653066, 0.60653066, 1., 0.36787944],
    >>>               [0.60653066, 0.60653066, 0.36787944, 1.]])
    >>> sample(K, 1.0, 10)
    [[0, 1, 1, 1],
     [0, 0, 0, 0],
     [1, 0, 0, 0],
     [0, 0, 0, 1],
     [0, 1, 1, 0],
     [2, 0, 0, 0],
     [0, 0, 0, 0],
     [0, 0, 0, 0],
     [0, 0, 1, 1],
     [0, 0, 0, 0]]

    Args:
        K (array): the positive semidefinite kernel matrix
        n_mean (float): average number of points per sample
        n_samples (int): number of samples to be generated

    Returns:
        samples (list[list[int]]): samples generated by the point process
    """
    ls, O = rescale_adjacency_matrix_thermal(K, n_mean)
    return np.array(generate_thermal_samples(ls, O, num_samples=n_samples)).tolist()



# Copyright 2019 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""
.. _Tensorflow_backend:

Tensorflow simulator backend
##############################

**Module name:** :mod:`strawberryfields.backends.tfbackend`

.. currentmodule:: strawberryfields.backends.tfbackend

The :class:`TFBackend` object implements a simulation of quantum optical circuits using
`Tensorflow <http://www.Tensorflow.org/>`_. The primary component of the TFBackend is a
:class:`Circuit` object which is used to simulate a multi-mode quantum optical system. The
:class:`TFBackend` provides the basic API-compatible interface to the simulator, while the
:class:`Circuit` object actually carries out the mathematical simulation.


The :class:`Circuit` simulator maintains an internal tensor representation of the quantum state of a multi-mode quantum optical system
using a (truncated) Fock basis representation. As its various state manipulation methods are called, the quantum state is updated
to reflect these changes. The simulator will try to keep the internal state in a pure (vector) representation
for as long as possible. Unitary gates will not change the type of representation, while state preparations and measurements will.

A number of factors determine the shape and dimensionality of the state tensor:

* the underlying state representation being used (either a ket vector or a density matrix)
* the number of modes :math:`n` actively being simulated
* the cutoff dimension :math:`D` for the Fock basis
* whether the circuit is operating in *batched mode* (with batch size :math:`B`)

When not operating in batched mode, the state tensor corresponds to a single multimode quantum system. If the
representation is a pure state, the state tensor has shape
:math:`(\underbrace{D,...,D}_{n~\text{times}})`.
In a mixed state representation, the state tensor has shape
:math:`(\underbrace{D,D,...,D,D}_{2n~\text{times}})`.
Indices for the same mode appear consecutively. Hence, for a mixed state, the first two indices
are for the first mode, the second are for the second mode, etc.

In batched mode, the state tensor simultaneously encodes an ensemble of :math:`B` multimode quantum systems
(indexed using the first axis of the state tensor). Pure states thus have shape
:math:`(B,\underbrace{D,...,D}_{n~\text{times}})`,
while mixed states have shape
:math:`(B,\underbrace{D,D,...,D,D}_{2n~\text{times}})`.

.. currentmodule:: strawberryfields.backends.tfbackend.TFBackend

TFBackend methods
=================

The parameters supplied for these operations can be either numeric (float, complex) values
or Tensorflow :class:`Variables`/:class:`Tensors`. The Tensorflow objects can either be scalars or vectors. For
vectors, they must have the same dimension as the declared batch size of the underlying circuit.

.. autosummary::
   supports
   begin_circuit
   add_mode
   del_mode
   get_modes
   get_cutoff_dim
   reset
   state
   is_vacuum
   prepare_vacuum_state
   prepare_coherent_state
   prepare_squeezed_state
   prepare_displaced_squeezed_state
   prepare_thermal_state
   prepare_fock_state
   prepare_ket_state
   prepare_dm_state
   rotation
   displacement
   squeeze
   beamsplitter
   cubic_phase
   kerr_interaction
   cross_kerr_interaction
   loss
   thermal_loss
   measure_homodyne
   measure_fock
   graph


Code details
~~~~~~~~~~~~
.. autoclass:: strawberryfields.backends.tfbackend.TFBackend
   :members:

FockStateTF
================

.. currentmodule:: strawberryfields.backends.tfbackend.states.FockStateTF

This class represents the quantum state
returned by the Tensorflow backend. It extends :class:`~.BaseFockState` with additional functionality
unique to the Tensorflow backend. The primary difference between this class and the Base Fock state
is that its methods and attributes can return either numerical or symbolic values.

The default representation (numerical or symbolic)
is set when creating the state: :code:`state = eng.run('backend', eval=True/False)`. The representation can also be specified on a per-use basis when calling a method, e.g., :code:`state.mean_photon(eval=True/False)`. Along with the boolean :code:`eval`, acceptable keyword arguments are :code:`session` (a Tensorflow Session object) and :code:`feed_dict` (a dictionary mapping Tensorflow objects to numerical values). These will be used when evaluating any Tensors.

.. currentmodule:: strawberryfields.backends.tfbackend.states.FockStateTF

.. autosummary::
   ket
   dm
   reduced_dm
   trace
   fock_prob
   all_fock_probs
   fidelity
   fidelity_coherent
   fidelity_vacuum
   is_vacuum
   quad_expectation
   mean_photon
   batched
   cutoff_dim
   graph


Code details
~~~~~~~~~~~~

.. autoclass:: strawberryfields.backends.tfbackend.states.FockStateTF
   :members:

"""
import sys

try:
    import tensorflow
except ImportError:
    tf_available = False
    tf_version = None
else:
    tf_available = True
    tf_version = tensorflow.__version__


tf_info = """\
To use Strawberry Fields with TensorFlow support, version 1.3 of TensorFlow is required.
This can be installed as follows:

pip install tensorflow==1.3
"""


tf_info_python = """\
To use Strawberry Fields with TensorFlow support, version 1.3 of TensorFlow is required.

Note that TensorFlow version 1.3 is only supported on Python versions
less than or equal to 3.6. To continue using TensorFlow with Strawberry Fields,
you will need to install Python 3.6.

The recommended method is to install Anaconda3:

https://www.anaconda.com/download

Once installed, you can then create a Python 3.6 Conda environment:

conda create --name sf_tensorflow_env python=3.6
conda activate sf_tensorflow_env
pip install strawberryfields tensorflow==1.3
"""


def excepthook(type, value, traceback):
    """Exception hook to suppress superfluous exceptions"""
    # pylint: disable=unused-argument
    print(value)


if not (tf_available and tf_version[:3] == "1.3"):
    sys.excepthook = excepthook

    if sys.version_info[1] > 6:
        raise ImportError(tf_info_python)

    raise ImportError(tf_info)


from .backend import TFBackend
from .ops import def_type as tf_complex_type



# Copyright 2019 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tensorflow states module"""
from string import ascii_lowercase as indices

import numpy as np
import tensorflow as tf
from scipy.special import factorial

from strawberryfields.backends.states import BaseFockState
from .ops import _check_for_eval, def_type, ladder_ops, phase_shifter_matrix, reduced_density_matrix


class FockStateTF(BaseFockState):
    r"""Class for the representation of quantum states in the Fock basis using the TFBackend.

    Args:
            state_data (array): the state representation in the Fock basis
            num_modes (int): the number of modes in the state
            pure (bool): True if the state is a pure state, false if the state is mixed
            cutoff_dim (int): the Fock basis truncation size
            mode_names (Sequence): (optional) this argument contains a list providing mode names
                    for each mode in the state.
            eval (bool): indicates the default return behaviour for the class instance (symbolic when eval=False, numerical when eval=True)
    """

    def __init__(
        self,
        state_data,
        num_modes,
        pure,
        cutoff_dim,
        graph,
        batched=False,
        mode_names=None,
        eval=True,
    ):
        # pylint: disable=too-many-arguments
        state_data = tf.convert_to_tensor(
            state_data, name="state_data"
        )  # convert everything to a Tensor so we have the option to do symbolic evaluations
        super().__init__(state_data, num_modes, pure, cutoff_dim, mode_names)
        self._batched = batched
        self._str = "<FockStateTF: num_modes={}, cutoff={}, pure={}, batched={}, hbar={}>".format(
            self.num_modes, self.cutoff_dim, self._pure, self._batched, self._hbar
        )
        self._graph = graph
        self._eval = eval

    def _run(self, t, **kwargs):
        """
        Numerically evaluates Tensor `t` based on the arguments in **kwargs.

        Args:
            t: Tensor (or iterable of Tensors) to (potentially) numerically evaluate.
            **kwargs: can be used to pass a session or a feed_dict. Otherwise a temporary session
                and no feed_dict will be used.

        Returns:
            The numerical evaluation of the Tensor t. Type is the same as the dtype of t
        """
        with self._graph.as_default():
            if "eval" not in kwargs:
                kwargs[
                    "eval"
                ] = (
                    self._eval
                )  # when the caller did not specify, use the default behaviour of this instance
            evaluate, session, feed_dict, close_session = _check_for_eval(kwargs)

            if evaluate:
                t = session.run(t, feed_dict=feed_dict)
                if close_session:
                    session.close()
            return t

    def trace(self, **kwargs):
        r"""
        Computes the trace of the state. May be numerical or symbolic.

        Args:
            **kwargs: Optional keyword arguments.

                * If this contains the key
                  ``eval``, then the corresponding argument will be used to determine the return behaviour of this function.
                  When ``eval=True``, the return value is numerical; when ``eval=False``, it is symbolic.
                * If eval is not present in kwargs, then state falls back to the an internal evaluation behaviour,
                  which is specified at initialization.
                * A Tensorflow Session or feed_dict may also be passed via the keys ``session`` or ``feed_dict``, respectively.
                  If a Session is supplied, then ``eval`` is overriden and the numerical evaluation takes place in the provided Session.
                  If session and/or feed_dict are not given, then a temporary session and/or empty feed_dict will be used.

        Returns:
            float/Tensor: the numerical value, or an unevaluated Tensor object, for the trace.
        """
        with self._graph.as_default():
            s = self.data
            if not self.batched:
                s = tf.expand_dims(s, 0)  # insert fake batch dimension
            if self.is_pure:
                flat_state = tf.reshape(s, [-1, self.cutoff_dim ** self.num_modes])
                tr = tf.reduce_sum(flat_state * tf.conj(flat_state), 1)
            else:
                # hack because tensorflow einsum doesn't allow partial trace
                tr = s
                for _ in range(self.num_modes):
                    tr = tf.trace(tr)
            if not self.batched:
                tr = tf.squeeze(tr, 0)  # drop fake batch dimension

            tr = self._run(tr, **kwargs)
            return tr

    def fock_prob(self, n, **kwargs):
        r"""
        Compute the probabilities of a specific Fock-basis matrix element for the state. May be numerical or symbolic.

        Args:
            n (Sequence[int]): the Fock state :math:`\ket{\vec{n}}` that we want to measure the probability of
            **kwargs: Optional keyword arguments.

                * If this contains the key
                  ``eval``, then the corresponding argument will be used to determine the return behaviour of this function.
                  When ``eval=True``, the return value is numerical; when ``eval=False``, it is symbolic.
                * If eval is not present in kwargs, then state falls back to the an internal evaluation behaviour,
                  which is specified at initialization.
                * A Tensorflow Session or feed_dict may also be passed via the keys ``session`` or ``feed_dict``, respectively.
                  If a Session is supplied, then ``eval`` is overriden and the numerical evaluation takes place in the provided Session.
                  If session and/or feed_dict are not given, then a temporary session and/or empty feed_dict will be used.
        Returns:
            float/Tensor: the numerical values, or an unevaluated Tensor object, for the Fock-state probabilities.
        """
        with self.graph.as_default():

            if len(n) != self.num_modes:
                raise ValueError("List length should be equal to number of modes")

            elif max(n) >= self.cutoff_dim:
                raise ValueError("Can't get distribution beyond truncation level")

            state = self.data

            if not self.batched:
                state = tf.expand_dims(state, 0)  # add fake batch dimension

            # put batch dimension at end to match behaviour of tf.gather_nd
            if self.is_pure:
                flat_idx = np.ravel_multi_index(n, [self.cutoff_dim] * self.num_modes)
                flat_state = tf.reshape(state, [-1, self.cutoff_dim ** self.num_modes])
                prob = tf.abs(flat_state[:, flat_idx]) ** 2
            else:
                doubled_n = [n[i // 2] for i in range(len(n) * 2)]
                flat_idx = np.ravel_multi_index(doubled_n, [self.cutoff_dim] * (2 * self.num_modes))
                flat_state = tf.reshape(state, [-1, self.cutoff_dim ** (2 * self.num_modes)])
                prob = flat_state[:, flat_idx]

            if not self.batched:
                prob = tf.squeeze(prob, 0)  # drop fake batch dimension

            prob = tf.real(prob)
            prob = tf.identity(prob, name="fock_prob")
            prob = self._run(prob, **kwargs)

            return prob

    def all_fock_probs(self, **kwargs):
        r"""
        Compute the probabilities of all possible Fock-basis states for the state. May be numerical or symbolic.

        For example, in the case of 3 modes, this method allows the Fock state probability
        :math:`|\braketD{0,2,3}{\psi}|^2` to be returned via

        .. code-block:: python

            probs = state.all_fock_probs()
            probs[0,2,3]

        Args:
            **kwargs: Optional keyword arguments.

                * If this contains the key ``eval``, then the corresponding argument will be used to determine the return behaviour of this function.
                  When ``eval=True``, the return value is numerical; when ``eval=False``, it is symbolic.
                * If eval is not present in kwargs, then state falls back to the an internal evaluation behaviour,
                  which is specified at initialization.
                * A Tensorflow Session or feed_dict may also be passed via the keys ``session`` or ``feed_dict``, respectively.
                  If a Session is supplied, then ``eval`` is overriden and the numerical evaluation takes place in the provided Session.
                  If session and/or feed_dict are not given, then a temporary session and/or empty feed_dict will be used.
        Returns:
            array/Tensor: the numerical values, or an unevaluated Tensor object, for the Fock-basis probabilities.
        """

        with self.graph.as_default():

            state = self.data
            if not self.batched:
                state = tf.expand_dims(state, 0)  # add fake batch dimension

            if self.is_pure:
                probs = tf.abs(state) ** 2
            else:
                # convert to index scheme compatible with tf.diag_part
                rng = np.arange(1, 2 * self.num_modes + 1, 2)
                perm = np.concatenate([[0], rng, rng + 1])
                perm_state = tf.transpose(state, perm)
                probs = tf.map_fn(
                    tf.diag_part, perm_state
                )  # diag_part doesn't work with batches, need to use map_fn

            if not self.batched:
                probs = tf.squeeze(probs, 0)  # drop fake batch dimension

            probs = tf.identity(probs, name="all_fock_probs")
            probs = self._run(probs, **kwargs)

            return probs

    def fidelity(self, other_state, mode, **kwargs):
        r"""
        Compute the fidelity of the reduced state (on the specified mode) with the state. May be numerical or symbolic.

        Args:
            other_state (array): state vector (ket) to compute the fidelity with respect to
            mode (int): which subsystem to use for the fidelity computation
            **kwargs: Optional keyword arguments.

                * If this contains the key
                  ``eval``, then the corresponding argument will be used to determine the return behaviour of this function.
                  When ``eval=True``, the return value is numerical; when ``eval=False``, it is symbolic.
                * If eval is not present in kwargs, then state falls back to the an internal evaluation behaviour,
                  which is specified at initialization.
                * A Tensorflow Session or feed_dict may also be passed via the keys ``session`` or ``feed_dict``, respectively.
                  If a Session is supplied, then ``eval`` is overriden and the numerical evaluation takes place in the provided Session.
                  If session and/or feed_dict are not given, then a temporary session and/or empty feed_dict will be used.
        Returns:
            float/Tensor: the numerical value, or an unevaluated Tensor object, for the fidelity.
        """
        with self.graph.as_default():
            rho = self.reduced_dm([mode])  # don't pass kwargs yet

            if not self.batched:
                rho = tf.expand_dims(rho, 0)  # add fake batch dimension
            if len(other_state.shape) == 1:
                other_state = tf.expand_dims(other_state, 0)  # add batch dimension for state

            other_state = tf.cast(other_state, def_type)
            state_dm = tf.einsum("bi,bj->bij", tf.conj(other_state), other_state)
            flat_state_dm = tf.reshape(state_dm, [1, -1])
            flat_rho = tf.reshape(rho, [-1, self.cutoff_dim ** 2])

            f = tf.real(
                tf.reduce_sum(flat_rho * flat_state_dm, axis=1)
            )  # implements a batched tr(rho|s><s|)

            if not self.batched:
                f = tf.squeeze(f, 0)  # drop fake batch dimension

            f = tf.identity(f, name="fidelity")
            f = self._run(f, **kwargs)

            return f

    def fidelity_coherent(self, alpha_list, **kwargs):
        r"""
        Compute the fidelity of the state with the coherent states specified by alpha_list. May be numerical or symbolic.

        Args:
            alpha_list (Sequence[complex]): list of coherence parameter values, one for each mode
            **kwargs: Optional keyword arguments.

                * If this contains the key
                  ``eval``, then the corresponding argument will be used to determine the return behaviour of this function.
                  When ``eval=True``, the return value is numerical; when ``eval=False``, it is symbolic.
                * If eval is not present in kwargs, then state falls back to the an internal evaluation behaviour,
                  which is specified at initialization.
                * A Tensorflow Session or feed_dict may also be passed via the keys ``session`` or ``feed_dict``, respectively.
                  If a Session is supplied, then ``eval`` is overriden and the numerical evaluation takes place in the provided Session.
                  If session and/or feed_dict are not given, then a temporary session and/or empty feed_dict will be used.
        Returns:
            float/Tensor: the numerical value, or an unevaluated Tensor object, for the fidelity :math:`\bra{\vec{\alpha}}\rho\ket{\vec{\alpha}}`.
        """
        with self.graph.as_default():
            if not hasattr(alpha_list, "__len__"):
                alpha_list = [alpha_list]

            if len(alpha_list) != self.num_modes:
                raise ValueError("The number of alpha values must match the number of modes.")

            max_indices = (len(indices) - 1) // 2
            if len(alpha_list) > max_indices:
                raise ValueError("Length of `alpha_list` exceeds supported number of modes.")

            s = self.data
            if not self.batched:
                s = tf.expand_dims(s, 0)  # introduce fake batch dimension

            coh = lambda a, dim: [
                np.exp(-0.5 * np.abs(a) ** 2) * (a) ** n / np.sqrt(factorial(n)) for n in range(dim)
            ]
            multi_cohs_list = [
                coh(a, self.cutoff_dim) for a in alpha_list
            ]  # shape is: [num_modes, cutoff_dim]
            eqn = ",".join(indices[: self._modes]) + "->" + indices[: self._modes]
            multi_cohs_vec = np.einsum(
                eqn, *multi_cohs_list
            )  # tensor product of specified coherent states
            flat_multi_cohs = np.reshape(
                multi_cohs_vec, [1, self.cutoff_dim ** self.num_modes]
            )  # flattened tensor product; shape is: [1, cutoff_dim * num_modes]

            if self.is_pure:
                flat_state = tf.reshape(s, [-1, self.cutoff_dim ** self.num_modes])
                ovlap = tf.reduce_sum(flat_multi_cohs.conj() * flat_state, axis=1)
                f = tf.abs(ovlap) ** 2
            else:
                batch_index = indices[0]
                free_indices = indices[1:]
                bra_indices = free_indices[: self.num_modes]
                ket_indices = free_indices[self.num_modes : 2 * self.num_modes]
                eqn = (
                    bra_indices
                    + ","
                    + batch_index
                    + "".join(bra_indices[idx] + ket_indices[idx] for idx in range(self.num_modes))
                    + ","
                    + ket_indices
                    + "->"
                    + batch_index
                )
                f = tf.einsum(
                    eqn,
                    tf.convert_to_tensor(np.conj(multi_cohs_vec), dtype=def_type),
                    s,
                    tf.convert_to_tensor(multi_cohs_vec, def_type),
                )
            if not self.batched:
                f = tf.squeeze(f, 0)  # drop fake batch dimension

            f = tf.identity(f, name="fidelity_coherent")
            f = self._run(f, **kwargs)

            return f

    def fidelity_vacuum(self, **kwargs):
        r"""
        Compute the fidelity of the state with the vacuum state. May be numerical or symbolic.

        Args:
            **kwargs: Optional keyword arguments.

                * If this contains the key
                  ``eval``, then the corresponding argument will be used to determine the return behaviour of this function.
                  When ``eval=True``, the return value is numerical; when ``eval=False``, it is symbolic.
                * If eval is not present in kwargs, then state falls back to the an internal evaluation behaviour,
                  which is specified at initialization.
                * A Tensorflow Session or feed_dict may also be passed via the keys ``session`` or ``feed_dict``, respectively.
                  If a Session is supplied, then ``eval`` is overriden and the numerical evaluation takes place in the provided Session.
                  If session and/or feed_dict are not given, then a temporary session and/or empty feed_dict will be used.

        Returns:
            float/Tensor: the numerical value, or an unevaluated Tensor object, for the fidelity :math:`\bra{\vec{0}}\rho\ket{\vec{0}}`.

        """
        with self._graph.as_default():
            mode_size = 1 if self.is_pure else 2
            if self.batched:
                vac_elem = tf.real(
                    tf.reshape(self.data, [-1, self.cutoff_dim ** (self.num_modes * mode_size)])[
                        :, 0
                    ]
                )
            else:
                vac_elem = tf.abs(tf.reshape(self.data, [-1])[0]) ** 2

            v = tf.identity(vac_elem, name="fidelity_vacuum")
            v = self._run(v, **kwargs)

            return v

    def is_vacuum(self, tol=0.0, **kwargs):
        r"""
        Computes a boolean which indicates whether the state is the vacuum state. May be numerical or symbolic.

        Args:
            tol: numerical tolerance. If the state has fidelity with vacuum within tol, then this method returns True.
            **kwargs: Optional keyword arguments.

                * If this contains the key
                  ``eval``, then the corresponding argument will be used to determine the return behaviour of this function.
                  When ``eval=True``, the return value is numerical; when ``eval=False``, it is symbolic.
                * If eval is not present in kwargs, then state falls back to the an internal evaluation behaviour,
                  which is specified at initialization.
                * A Tensorflow Session or feed_dict may also be passed via the keys ``session`` or ``feed_dict``, respectively.
                  If a Session is supplied, then ``eval`` is overriden and the numerical evaluation takes place in the provided Session.
                  If session and/or feed_dict are not given, then a temporary session and/or empty feed_dict will be used.

        Returns:
            bool/Tensor: the boolean value, or an unevaluated Tensor object, for whether the state is in vacuum.
        """
        with self._graph.as_default():
            fidel = self.fidelity_vacuum()  # dont pass on kwargs yet
            is_vac = tf.less_equal(1 - fidel, tol, name="is_vacuum")
            is_vac = self._run(is_vac, **kwargs)
            return is_vac

    def reduced_dm(self, modes, **kwargs):
        r"""
        Computes the reduced density matrix representation of the state. May be numerical or symbolic.

        Args:
            modes (int or Sequence[int]): specifies the mode(s) to return the reduced
                                density matrix for.
            **kwargs: Optional keyword arguments.

                * If this contains the key
                  ``eval``, then the corresponding argument will be used to determine the return behaviour of this function.
                  When ``eval=True``, the return value is numerical; when ``eval=False``, it is symbolic.
                * If eval is not present in kwargs, then state falls back to the an internal evaluation behaviour,
                  which is specified at initialization.
                * A Tensorflow Session or feed_dict may also be passed via the keys ``session`` or ``feed_dict``, respectively.
                  If a Session is supplied, then ``eval`` is overriden and the numerical evaluation takes place in the provided Session.
                  If session and/or feed_dict are not given, then a temporary session and/or empty feed_dict will be used.

        Returns:
            array/Tensor: the numerical value, or an unevaluated Tensor object, for the density matrix.
        """
        with self._graph.as_default():
            if modes == list(range(self.num_modes)):
                # reduced state is full state
                return self.dm(**kwargs)

            if isinstance(modes, int):
                modes = [modes]
            if modes != sorted(modes):
                raise ValueError("The specified modes cannot be duplicated.")
            if len(modes) > self.num_modes:
                raise ValueError(
                    "The number of specified modes cannot "
                    "be larger than the number of subsystems."
                )

            reduced = self.dm(**kwargs)
            for m in modes:
                reduced = reduced_density_matrix(reduced, m, False, batched=self.batched)

            s = tf.identity(reduced, name="reduced_density_matrix")
            s = self._run(s, **kwargs)
            return s

    def quad_expectation(self, mode, phi=0.0, **kwargs):
        r"""
        Compute the expectation value of the quadrature operator :math:`\hat{x}_\phi` for the reduced state on the specified mode. May be numerical or symbolic.

        Args:
            mode (int): which subsystem to take the expectation value of
            phi (float): rotation angle for the quadrature operator
            **kwargs: Optional keyword arguments.

                * If this contains the key
                  ``eval``, then the corresponding argument will be used to determine the return behaviour of this function.
                  When ``eval=True``, the return value is numerical; when ``eval=False``, it is symbolic.
                * If eval is not present in kwargs, then state falls back to the an internal evaluation behaviour,
                  which is specified at initialization.
                * A Tensorflow Session or feed_dict may also be passed via the keys ``session`` or ``feed_dict``, respectively.
                  If a Session is supplied, then ``eval`` is overriden and the numerical evaluation takes place in the provided Session.
                  If session and/or feed_dict are not given, then a temporary session and/or empty feed_dict will be used.

        Returns:
                float/Tensor: the numerical value, or an unevaluated Tensor object, for the expectation value
        """
        with self.graph.as_default():
            rho = self.reduced_dm([mode])  # don't pass kwargs yet

            phi = tf.convert_to_tensor(phi)
            if self.batched and len(phi.shape) == 0:  # pylint: disable=len-as-condition
                phi = tf.expand_dims(phi, 0)
            larger_cutoff = (
                self.cutoff_dim + 1
            )  # start one dimension higher to avoid truncation errors
            R = phase_shifter_matrix(phi, larger_cutoff, batched=self.batched)
            if not self.batched:
                R = tf.expand_dims(R, 0)  # add fake batch dimension

            a, ad = ladder_ops(larger_cutoff)
            x = np.sqrt(self._hbar / 2.0) * (a + ad)
            x = tf.expand_dims(tf.cast(x, def_type), 0)  # add batch dimension to x
            quad = tf.conj(R) @ x @ R
            quad2 = (quad @ quad)[:, : self.cutoff_dim, : self.cutoff_dim]
            quad = quad[
                :, : self.cutoff_dim, : self.cutoff_dim
            ]  # drop highest dimension; remaining array gives correct truncated x**2

            if not self.batched:
                rho = tf.expand_dims(rho, 0)  # add fake batch dimension

            flat_rho = tf.reshape(rho, [-1, self.cutoff_dim ** 2])
            flat_quad = tf.reshape(quad, [1, self.cutoff_dim ** 2])
            flat_quad2 = tf.reshape(quad2, [1, self.cutoff_dim ** 2])

            e = tf.real(
                tf.reduce_sum(flat_rho * flat_quad, axis=1)
            )  # implements a batched tr(rho * x)
            e2 = tf.real(
                tf.reduce_sum(flat_rho * flat_quad2, axis=1)
            )  # implements a batched tr(rho * x ** 2)
            v = e2 - e ** 2

            if not self.batched:
                e = tf.squeeze(e, 0)  # drop fake batch dimension
                v = tf.squeeze(v, 0)  # drop fake batch dimension

            e = tf.identity(e, name="quad_expectation")
            v = tf.identity(v, name="quad_variance")
            e, v = self._run([e, v], **kwargs)

            return e, v

    def mean_photon(self, mode, **kwargs):
        r"""
        Compute the mean photon number for the reduced state on the specified mode. May be numerical or symbolic.

        Args:
            mode (int): which subsystem to take the mean photon number of
            **kwargs: Optional keyword arguments.

                * If this contains the key
                  ``eval``, then the corresponding argument will be used to determine the return behaviour of this function.
                  When ``eval=True``, the return value is numerical; when ``eval=False``, it is symbolic.
                * If eval is not present in kwargs, then state falls back to the an internal evaluation behaviour,
                  which is specified at initialization.
                * A Tensorflow Session or feed_dict may also be passed via the keys ``session`` or ``feed_dict``, respectively.
                  If a Session is supplied, then ``eval`` is overriden and the numerical evaluation takes place in the provided Session.
                  If session and/or feed_dict are not given, then a temporary session and/or empty feed_dict will be used.
        Returns:
            tuple(float/Tensor): tuple containing the numerical value, or an unevaluated Tensor object, for the mean photon number and variance.
        """
        with self.graph.as_default():
            rho = self.reduced_dm([mode])  # don't pass kwargs yet

            if not self.batched:
                rho = tf.expand_dims(rho, 0)  # add fake batch dimension

            n = np.diag(np.arange(self.cutoff_dim))
            flat_n = np.reshape(n, [1, self.cutoff_dim ** 2])

            flat_rho = tf.reshape(rho, [-1, self.cutoff_dim ** 2])

            nbar = tf.real(
                tf.reduce_sum(flat_rho * flat_n, axis=1)
            )  # implements a batched tr(rho * n)
            nbarSq = tf.real(
                tf.reduce_sum(flat_rho * flat_n ** 2, axis=1)
            )  # implements a batched tr(rho * n^2)
            var = nbarSq - nbar ** 2

            if not self.batched:
                nbar = tf.squeeze(nbar, 0)  # drop fake batch dimension
                var = tf.squeeze(var, 0)  # drop fake batch dimension

            nbar = tf.identity(nbar, name="mean_photon")
            var = tf.identity(var, name="mean_photon_variance")

            nbar = self._run(nbar, **kwargs)
            var = self._run(var, **kwargs)

            return nbar, var

    def ket(self, **kwargs):
        r"""
        Computes the ket representation of the state. May be numerical or symbolic.

        Args:
            **kwargs: Optional keyword arguments.

                * If this contains the key
                  ``eval``, then the corresponding argument will be used to determine the return behaviour of this function.
                  When ``eval=True``, the return value is numerical; when ``eval=False``, it is symbolic.
                * If eval is not present in kwargs, then state falls back to the an internal evaluation behaviour,
                  which is specified at initialization.
                * A Tensorflow Session or feed_dict may also be passed via the keys ``session`` or ``feed_dict``, respectively.
                  If a Session is supplied, then ``eval`` is overriden and the numerical evaluation takes place in the provided Session.
                  If session and/or feed_dict are not given, then a temporary session and/or empty feed_dict will be used.

        Returns:
            array/Tensor: the numerical value, or an unevaluated Tensor object, for the ket.
        """
        with self._graph.as_default():
            if not self.is_pure:
                return None

            s = tf.identity(self.data, name="ket")
            s = self._run(s, **kwargs)
            return s

    def dm(self, **kwargs):
        r"""
        Computes the density matrix representation of the state. May be numerical or symbolic.

        Args:
            **kwargs: Optional keyword arguments.

                * If this contains the key
                  ``eval``, then the corresponding argument will be used to determine the return behaviour of this function.
                  When ``eval=True``, the return value is numerical; when ``eval=False``, it is symbolic.
                * If eval is not present in kwargs, then state falls back to the an internal evaluation behaviour,
                  which is specified at initialization.
                * A Tensorflow Session or feed_dict may also be passed via the keys ``session`` or ``feed_dict``, respectively.
                  If a Session is supplied, then ``eval`` is overriden and the numerical evaluation takes place in the provided Session.
                  If session and/or feed_dict are not given, then a temporary session and/or empty feed_dict will be used.

        Returns:
            array/Tensor: the numerical value, or an unevaluated Tensor object, for the density matrix.
        """
        with self._graph.as_default():
            if self.is_pure:
                if self.batched:
                    batch_index = indices[0]
                    free_indices = indices[1:]
                else:
                    batch_index = ""
                    free_indices = indices
                ket = self.data
                left_str = [batch_index] + [
                    free_indices[i] for i in range(0, 2 * self.num_modes, 2)
                ]
                right_str = [batch_index] + [
                    free_indices[i] for i in range(1, 2 * self.num_modes, 2)
                ]
                out_str = [batch_index] + [free_indices[: 2 * self.num_modes]]
                einstr = "".join(left_str + [","] + right_str + ["->"] + out_str)
                s = tf.einsum(einstr, ket, tf.conj(ket))
            else:
                s = tf.identity(self.data, name="density_matrix")
            s = self._run(s, **kwargs)
            return s

    def wigner(self, mode, xvec, pvec):
        r"""Calculates the discretized Wigner function of the specified mode.

        .. warning::

            Calculation of the Wigner function is currently only supported if
            ``eval=True`` and ``batched=False``.

        .. note::

            This code is a modified version of the 'iterative' method of the
            `wigner function provided in QuTiP <http://qutip.org/docs/4.0.2/apidoc/functions.html?highlight=wigner#qutip.wigner.wigner>`_,
            which is released under the BSD license, with the following
            copyright notice:

            Copyright (C) 2011 and later, P.D. Nation, J.R. Johansson,
            A.J.G. Pitchford, C. Granade, and A.L. Grimsmo. All rights reserved.

        Args:
            mode (int): the mode to calculate the Wigner function for
            xvec (array): array of discretized :math:`x` quadrature values
            pvec (array): array of discretized :math:`p` quadrature values

        Returns:
            array: 2D array of size [len(xvec), len(pvec)], containing reduced Wigner function
            values for specified x and p values.
        """
        if self._eval and not self.batched:
            return super().wigner(mode, xvec, pvec)

        raise NotImplementedError(
            "Calculation of the Wigner function is currently "
            "only supported when eval=True and batched=False"
        )

    def poly_quad_expectation(self, A, d=None, k=0, phi=0, **kwargs):  # pragma: no cover
        r"""The multi-mode expectation values and variance of arbitrary 2nd order polynomials
        of quadrature operators.

        .. warning::

            Calculation of multi-mode quadratic expectation values is currently only supported if
            ``eval=True`` and ``batched=False``.

        An arbitrary 2nd order polynomial of quadrature operators over $N$ modes can always
        be written in the following form:

        .. math:: P(\mathbf{r}) = \mathbf{r}^T A\mathbf{r} + \mathbf{r}^T \mathbf{d} + k I

        where:

        * :math:`A\in\mathbb{R}^{2N\times 2N}` is a symmetric matrix
          representing the quadratic coefficients,
        * :math:`\mathbf{d}\in\mathbb{R}^{2N}` is a real vector representing
          the linear coefficients,
        * :math:`k\in\mathbb{R}` represents the constant term, and
        * :math:`\mathbf{r} = (\x_1,\dots,\x_N,\p_1,\dots,\p_N)` is the vector
          of quadrature operators in :math:`xp`-ordering.

        This method returns the expectation value of this second-order polynomial,

        .. math:: \langle P(\mathbf{r})\rangle,

        as well as the variance

        .. math:: \Delta P(\mathbf{r})^2 = \langle P(\mathbf{r})^2\rangle - \braket{P(\mathbf{r})}^2

        Args:
            A (array): a real symmetric 2Nx2N NumPy array, representing the quadratic
                coefficients of the second order quadrature polynomial.
            d (array): a symmetric length-2N NumPy array, representing the linear
                coefficients of the second order quadrature polynomial. Defaults to the zero vector.
            k (float): the constant term. Default 0.
            phi (float): quadrature angle, clockwise from the positive :math:`x` axis. If provided,
                the vector of quadrature operators :math:`\mathbf{r}` is first rotated
                by angle :math:`\phi` in the phase space.

            **kwargs: Optional keyword arguments.

                * If this contains the key
                  ``eval``, then the corresponding argument will be used to determine the return behaviour of this function.
                  When ``eval=True``, the return value is numerical; when ``eval=False``, it is symbolic.
                * If eval is not present in kwargs, then state falls back to the an internal evaluation behaviour,
                  which is specified at initialization.
                * A Tensorflow Session or feed_dict may also be passed via the keys ``session`` or ``feed_dict``, respectively.
                  If a Session is supplied, then ``eval`` is overriden and the numerical evaluation takes place in the provided Session.
                  If session and/or feed_dict are not given, then a temporary session and/or empty feed_dict will be used.


        Returns:
            tuple (float, float): expectation value and variance
        """
        with self._graph.as_default():
            if "eval" not in kwargs:
                kwargs[
                    "eval"
                ] = (
                    self._eval
                )  # when the caller did not specify, use the default behaviour of this instance
            evaluate, session, feed_dict, close_session = _check_for_eval(kwargs)

            if evaluate and not self.batched:
                self._data = session.run(self._data, feed_dict=feed_dict)
                if close_session:
                    session.close()
                return super().poly_quad_expectation(A, d, k, phi, **kwargs)
            else:
                raise NotImplementedError(
                    "Calculation of multi-mode quadratic expectation values is currently "
                    "only supported when batched=False and when evaluating numerically."
                )

    @property
    def batched(self):
        """The number of batches."""
        return self._batched

    @property
    def graph(self):
        """The computational graph."""
        return self._graph



#!/usr/bin/env python3
import strawberryfields as sf
from strawberryfields.ops import *

# initialize engine and program objects
eng = sf.Engine(backend="fock", backend_options={"cutoff_dim": 7})
boson_sampling = sf.Program(4)

with boson_sampling.context as q:
    # prepare the input fock states
    Fock(1) | q[0]
    Fock(1) | q[1]
    Vac     | q[2]
    Fock(1) | q[3]

    # rotation gates
    Rgate(0.5719) | q[0]
    Rgate(-1.9782) | q[1]
    Rgate(2.0603) | q[2]
    Rgate(0.0644) | q[3]

    # beamsplitter array
    BSgate(0.7804, 0.8578)  | (q[0], q[1])
    BSgate(0.06406, 0.5165) | (q[2], q[3])
    BSgate(0.473, 0.1176)   | (q[1], q[2])
    BSgate(0.563, 0.1517)   | (q[0], q[1])
    BSgate(0.1323, 0.9946)  | (q[2], q[3])
    BSgate(0.311, 0.3231)   | (q[1], q[2])
    BSgate(0.4348, 0.0798)  | (q[0], q[1])
    BSgate(0.4368, 0.6157)  | (q[2], q[3])
    # end circuit

# run the engine
results = eng.run(boson_sampling)

# extract the joint Fock probabilities
probs = results.state.all_fock_probs()

# print the joint Fock state probabilities
print(probs[1, 1, 0, 1])
print(probs[2, 0, 0, 1])



# Copyright 2019 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# pylint: disable=too-many-public-methods
"""Gaussian backend"""
import warnings

from numpy import (
    empty,
    concatenate,
    array,
    identity,
    arctan2,
    angle,
    sqrt,
    dot,
    vstack,
    zeros_like,
    allclose,
    ix_,
)
from numpy.linalg import inv
from thewalrus.samples import hafnian_sample_state, torontonian_sample_state

from strawberryfields.backends import BaseGaussian
from strawberryfields.backends.shared_ops import changebasis

from .ops import xmat
from .gaussiancircuit import GaussianModes
from .states import GaussianState


class GaussianBackend(BaseGaussian):
    r"""The GaussianBackend implements a simulation of quantum optical circuits
    in NumPy using the Gaussian formalism, returning a :class:`~.GaussianState`
    state object.

    The primary component of the GaussianBackend is a
    :attr:`~.GaussianModes` object which is used to simulate a multi-mode quantum optical system.
    :class:`~.GaussianBackend` provides the basic API-compatible interface to the simulator, while the
    :attr:`~.GaussianModes` object actually carries out the mathematical simulation.

    The :attr:`GaussianModes` simulators maintain an internal covariance matrix & vector of means
    representation of a multi-mode quantum optical system.

    Note that unlike commonly used covariance matrix representations we encode our state in two complex
    matrices :math:`N` and :math:`M` that are defined as follows
    :math:`N_{i,j} = \langle a^\dagger _i a_j \rangle`
    :math:`M_{i,j} = \langle a _i a_j \rangle`
    and a vector of means :math:`\alpha_i =\langle a_i \rangle`.

    .. 
        .. currentmodule:: strawberryfields.backends.gaussianbackend
        .. autosummary::
            :toctree: api

            ~gaussiancircuit.GaussianModes
            ~ops
    """

    short_name = "gaussian"
    circuit_spec = "gaussian"

    def __init__(self):
        """Initialize the backend.
        """
        super().__init__()
        self._supported["mixed_states"] = True
        self._init_modes = None
        self.circuit = None

    def begin_circuit(self, num_subsystems, **kwargs):
        self._init_modes = num_subsystems
        self.circuit = GaussianModes(num_subsystems)

    def add_mode(self, n=1):
        self.circuit.add_mode(n)

    def del_mode(self, modes):
        self.circuit.del_mode(modes)

    def get_modes(self):
        return self.circuit.get_modes()

    def reset(self, pure=True, **kwargs):
        self.circuit.reset(self._init_modes)

    def prepare_thermal_state(self, nbar, mode):
        self.circuit.init_thermal(nbar, mode)

    def prepare_vacuum_state(self, mode):
        self.circuit.loss(0.0, mode)

    def prepare_coherent_state(self, alpha, mode):
        self.circuit.loss(0.0, mode)
        self.circuit.displace(alpha, mode)

    def prepare_squeezed_state(self, r, phi, mode):
        self.circuit.loss(0.0, mode)
        self.circuit.squeeze(r, phi, mode)

    def prepare_displaced_squeezed_state(self, alpha, r, phi, mode):
        self.circuit.loss(0.0, mode)
        self.circuit.squeeze(r, phi, mode)
        self.circuit.displace(alpha, mode)

    def rotation(self, phi, mode):
        self.circuit.phase_shift(phi, mode)

    def displacement(self, alpha, mode):
        self.circuit.displace(alpha, mode)

    def squeeze(self, z, mode):
        phi = angle(z)
        r = abs(z)
        self.circuit.squeeze(r, phi, mode)

    def beamsplitter(self, t, r, mode1, mode2):
        if isinstance(t, complex):
            raise ValueError("Beamsplitter transmittivity t must be a float.")
        theta = arctan2(abs(r), t)
        phi = angle(r)
        self.circuit.beamsplitter(-theta, -phi, mode1, mode2)

    def measure_homodyne(self, phi, mode, shots=1, select=None, **kwargs):
        r"""Measure a :ref:`phase space quadrature <homodyne>` of the given mode.

        See :meth:`.BaseBackend.measure_homodyne`.

        Keyword Args:
            eps (float): Homodyne amounts to projection onto a quadrature eigenstate.
                This eigenstate is approximated by a squeezed state whose variance has been
                squeezed to the amount ``eps``, :math:`V_\text{meas} = \texttt{eps}^2`.
                Perfect homodyning is obtained when ``eps`` :math:`\to 0`.

        Returns:
            float: measured value
        """
        if shots != 1:
            if select is not None:
                raise NotImplementedError(
                    "Gaussian backend currently does not support "
                    "postselection if shots != 1 for homodyne measurement"
                )

            raise NotImplementedError(
                "Gaussian backend currently does not support " "shots != 1 for homodyne measurement"
            )

        # phi is the rotation of the measurement operator, hence the minus
        self.circuit.phase_shift(-phi, mode)

        if select is None:
            qs = self.circuit.homodyne(mode, **kwargs)[0, 0]
        else:
            val = select * 2 / sqrt(2 * self.circuit.hbar)
            qs = self.circuit.post_select_homodyne(mode, val, **kwargs)

        return qs * sqrt(2 * self.circuit.hbar) / 2

    def measure_heterodyne(self, mode, shots=1, select=None):

        if shots != 1:
            if select is not None:
                raise NotImplementedError(
                    "Gaussian backend currently does not support "
                    "postselection if shots != 1 for heterodyne measurement"
                )

            raise NotImplementedError(
                "Gaussian backend currently does not support "
                "shots != 1 for heterodyne measurement"
            )

        if select is None:
            m = identity(2)
            res = 0.5 * self.circuit.measure_dyne(m, [mode], shots=shots)
            return res[0, 0] + 1j * res[0, 1]

        res = select
        self.circuit.post_select_heterodyne(mode, select)

        return res

    def prepare_gaussian_state(self, r, V, modes):
        if isinstance(modes, int):
            modes = [modes]

        # make sure number of modes matches shape of r and V
        N = len(modes)
        if len(r) != 2 * N:
            raise ValueError("Length of means vector must be twice the number of modes.")
        if V.shape != (2 * N, 2 * N):
            raise ValueError(
                "Shape of covariance matrix must be [2N, 2N], where N is the number of modes."
            )

        # convert xp-ordering to symmetric ordering
        means = vstack([r[:N], r[N:]]).reshape(-1, order="F")
        C = changebasis(N)
        cov = C @ V @ C.T

        self.circuit.fromscovmat(cov, modes)
        self.circuit.fromsmean(means, modes)

    def is_vacuum(self, tol=0.0, **kwargs):
        return self.circuit.is_vacuum(tol)

    def loss(self, T, mode):
        self.circuit.loss(T, mode)

    def thermal_loss(self, T, nbar, mode):
        self.circuit.thermal_loss(T, nbar, mode)

    def measure_fock(self, modes, shots=1, select=None):
        if shots != 1:
            if select is not None:
                raise NotImplementedError(
                    "Gaussian backend currently does not support " "postselection"
                )
            warnings.warn(
                "Cannot simulate non-Gaussian states. "
                "Conditional state after Fock measurement has not been updated."
            )

        mu = self.circuit.mean
        mean = self.circuit.smeanxp()
        cov = self.circuit.scovmatxp()

        x_idxs = array(modes)
        p_idxs = x_idxs + len(mu)
        modes_idxs = concatenate([x_idxs, p_idxs])
        reduced_cov = cov[ix_(modes_idxs, modes_idxs)]
        reduced_mean = mean[modes_idxs]
        # check we are sampling from a gaussian state with zero mean
        if allclose(mu, zeros_like(mu)):
            samples = hafnian_sample_state(reduced_cov, shots)
        else:
            samples = hafnian_sample_state(reduced_cov, shots, mean=reduced_mean)
        # for backward compatibility with previous measurement behaviour,
        # if only one shot, then we drop the shots axis
        if shots == 1:
            samples = samples.reshape((len(modes),))
        return samples

    def measure_threshold(self, modes, shots=1, select=None):
        if shots != 1:
            if select is not None:
                raise NotImplementedError(
                    "Gaussian backend currently does not support " "postselection"
                )
            warnings.warn(
                "Cannot simulate non-Gaussian states. "
                "Conditional state after Threshold measurement has not been updated."
            )

        mu = self.circuit.mean
        cov = self.circuit.scovmatxp()
        # check we are sampling from a gaussian state with zero mean
        if not allclose(mu, zeros_like(mu)):
            raise NotImplementedError(
                "Threshold measurement is only supported for " "Gaussian states with zero mean"
            )
        x_idxs = array(modes)
        p_idxs = x_idxs + len(mu)
        modes_idxs = concatenate([x_idxs, p_idxs])
        reduced_cov = cov[ix_(modes_idxs, modes_idxs)]
        samples = torontonian_sample_state(reduced_cov, shots)
        # for backward compatibility with previous measurement behaviour,
        # if only one shot, then we drop the shots axis
        if shots == 1:
            samples = samples.reshape((len(modes),))
        return samples

    def state(self, modes=None, **kwargs):
        """Returns the state of the quantum simulation.

        See :meth:`.BaseBackend.state`.

        Returns:
            GaussianState: state description
        """
        m = self.circuit.scovmat()
        r = self.circuit.smean()

        if modes is None:
            modes = list(range(len(self.get_modes())))

        listmodes = list(concatenate((2 * array(modes), 2 * array(modes) + 1)))
        covmat = empty((2 * len(modes), 2 * len(modes)))
        means = r[listmodes]

        for i, ii in enumerate(listmodes):
            for j, jj in enumerate(listmodes):
                covmat[i, j] = m[ii, jj]

        means *= sqrt(2 * self.circuit.hbar) / 2
        covmat *= self.circuit.hbar / 2

        mode_names = ["q[{}]".format(i) for i in array(self.get_modes())[modes]]

        # qmat and amat
        qmat = self.circuit.qmat()
        N = qmat.shape[0] // 2

        # work out if qmat and Amat need to be reduced
        if 1 <= len(modes) < N:
            # reduce qmat
            ind = concatenate([array(modes), N + array(modes)])
            rows = ind.reshape((-1, 1))
            cols = ind.reshape((1, -1))
            qmat = qmat[rows, cols]

            # calculate reduced Amat
            N = qmat.shape[0] // 2
            Amat = dot(xmat(N), identity(2 * N) - inv(qmat))
        else:
            Amat = self.circuit.Amat()

        return GaussianState((means, covmat), len(modes), qmat, Amat, mode_names=mode_names)
